from pytestqt import qtbot
import pytest
import os


# NOTE: Disabled on travis due to unknown issues

@pytest.mark.skipif(os.environ.get("TRAVIS", False),
                    reason="Thread tests don't work on travis for unknown reason; temporarily disabled")
def test_threads(qtbot):
    from xicam.core import threads
    from qtpy.QtCore import QObject, Signal


    def callback(a):
        assert a == 10

    t = threads.QThreadFuture(sum, [1, 2, 3, 4], callback_slot=callback)

    class Callback(QObject):
        sig = Signal(int)

    callback = Callback()
    t2 = threads.QThreadFuture(sum, [1, 2, 3, 4], callback_slot=callback.sig)

    t.start()
    t2.start()

    qtbot.waitSignals([t.sigFinished, t2.sigFinished])

@pytest.mark.skipif(os.environ.get("TRAVIS", False),
                    reason="Thread tests don't work on travis for unknown reason; temporarily disabled")
def test_threads_iterator(qtbot):
    from xicam.core import threads

    results = []

    def callback(a):
        results.append(a)

    def testiterator():
        for i in range(3):
            yield i

    def check():
        assert sum(results) == 3

    t = threads.QThreadFutureIterator(testiterator, callback_slot=callback, finished_slot=check)
    t.start()
    qtbot.waitSignal(t.sigFinished)




"""This module provides application-wide logging tools.

Unhandled exceptions are hooked into the log. Messages and progress
can be displayed in the main Xi-cam window using showProgress and showMessage.

Constants
---------
DEFAULT_FILE_LOG_LEVEL : int
    Default logging level for logging to file.
DEFAULT_STREAM_LOG_LEVEL : int
    Default logging level for logging to stream (e.g. sys.stdout).
FILE_LOG_LEVEL_SETTINGS_NAME : str
    Name of the settings value for defining the file logging level.
LOGGING_SETTINGS_NAME : str
    Name of the serialized settings via LogginSettingsPlugin (in GUI mode).
STREAM_LOG_LEVEL_SETTINGS_NAME : str
    Name of the settings value for defining the stream logging level.

"""
import cloudpickle as pickle
import logging
import faulthandler
import sys
import os
import time
import warnings
from typing import Any
import traceback
from collections import defaultdict
from qtpy.QtCore import QSettings, QTimer
from xicam.core import paths
from contextlib import contextmanager


# TODO: Add logging for images
# TODO: Add icons in GUI reflection

# Log levels constants
DEBUG = logging.DEBUG  # 10
INFO = logging.INFO  # 20
WARNING = logging.WARNING  # 30
ERROR = logging.ERROR  # 40
CRITICAL = logging.CRITICAL  # 50

levels = {DEBUG: "DEBUG", INFO: "INFO", WARNING: "WARNING", ERROR: "ERROR", CRITICAL: "CRITICAL"}

# Define constants that the LoggingSettingsPlugin will end up using as well
DEFAULT_FILE_LOG_LEVEL = DEBUG
DEFAULT_STREAM_LOG_LEVEL = WARNING
FILE_LOG_LEVEL_SETTINGS_NAME = "File Log Level"
LOGGING_SETTINGS_NAME = "Logging"
STREAM_LOG_LEVEL_SETTINGS_NAME = "Terminal Log Level"

# GUI widgets are registered into these slots to display messages/progress
statusbar = None
progressbar = None

# Create a log file that captures all logs (DEBUG)
log_dir = os.path.join(paths.user_cache_dir, "logs")
os.makedirs(log_dir, exist_ok=True)
log_file = "out.log"
logger = logging.getLogger("xicam")
logger.setLevel("DEBUG")  # minimum level shown

# Create a formatter that all handlers below can use for formatting their log messages
# format = "%(asctime)s - %(name)s - %(module)s:%(lineno)d - %(funcName)s - "
format = "%(asctime)s - %(caller_name)s - %(levelname)s - %(threadName)s > %(message)s"
date_format = "%a %b %d %H:%M:%S %Y"
formatter = logging.Formatter(fmt=format, datefmt=date_format)

# Use the QSettings set up for us by LoggingSettingsPlugin
logging_settings = QSettings().value(LOGGING_SETTINGS_NAME, {})
if logging_settings:
    # Unserialize the bytes stream (gives us a dict)
    logging_settings = pickle.loads(logging_settings)
    # Since LoggingSettingsPlugin is a ParameterSettingsPlugin,
    # we access the settings (parameters) through the "children" key
    logging_settings = logging_settings.get("children", {})

# Create a file handler for logging to a file
file_log_level = DEFAULT_FILE_LOG_LEVEL
# See if we have the file logging level available to us
file_log_settings = logging_settings.get(FILE_LOG_LEVEL_SETTINGS_NAME, {}).get("value")
if logging_settings and file_log_settings:
    file_log_level = file_log_settings
# By default, append to log file
file_handler = logging.FileHandler(os.path.join(log_dir, log_file))
file_handler.setLevel(file_log_level)  # minimum level shown
file_handler.setFormatter(formatter)
logger.addHandler(file_handler)

# Create a stream handler (attaches to sys.stderr by default)
stream_log_level = DEFAULT_STREAM_LOG_LEVEL
# See if we have the stream logging level available to us
stream_log_settings = logging_settings.get(STREAM_LOG_LEVEL_SETTINGS_NAME, {}).get("value")
if logging_settings and stream_log_settings:
    stream_log_level = stream_log_settings
stream_handler = logging.StreamHandler()
stream_handler.setLevel(stream_log_level)  # minimum level shown
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

trayicon = None
if "qtpy" in sys.modules:
    from qtpy.QtWidgets import QApplication

    if QApplication.instance():
        from qtpy.QtWidgets import QSystemTrayIcon
        from qtpy.QtGui import QIcon, QPixmap
        from xicam.gui.static import path

        trayicon = QSystemTrayIcon(QIcon(QPixmap(str(path("icons/cpu.png")))))  # TODO: better icon

_thread_count = 0


def _increment_thread():
    global _thread_count
    _thread_count += 1
    return _thread_count


threadIds = defaultdict(_increment_thread)


def showProgress(value: int, minval: int = 0, maxval: int = 100):
    """
    Displays the progress value on the subscribed QProgressBar (set as the global progressbar)

    Parameters
    ----------
    value   : int
        Progress value.
    minval  : int
        Minimum value (default: 0)
    maxval  : int
        Maximum value (default: 100)

    """
    if progressbar:
        from .. import threads  # must be a late import

        threads.invoke_in_main_thread(progressbar.show)
        threads.invoke_in_main_thread(progressbar.setRange, minval, maxval)
        threads.invoke_in_main_thread(progressbar.setValue, value)


def showBusy():
    """
     Displays a busy indicator on the subscribed QProgressBar (set as the global progressbar)

    """
    if progressbar:
        from .. import threads  # must be a late import

        threads.invoke_in_main_thread(progressbar.show)
        threads.invoke_in_main_thread(progressbar.setRange, 0, 0)


def hideBusy():
    """
    Stops a busy indicator on the subscribed QProgressBar (set as the global progressbar)

    """
    if progressbar:
        from .. import threads  # must be a late import

        threads.invoke_in_main_thread(progressbar.hide)
        threads.invoke_in_main_thread(progressbar.setRange, 0, 100)


# aliases
showReady = hideBusy
hideProgress = hideBusy


def notifyMessage(*args, timeout=8000, title="", level: int = INFO):
    """
    Same as logMessage, but displays to the subscribed notification system with a timeout.

    Parameters
    ----------
    args        :   tuple(str)
        See logMessage...
    timeout     :   int
        How long the message is displayed. If set 0, the message is persistent.
    kwargs      :   dict
        See logMessage...
    Returns
    -------

    """
    global trayicon
    if trayicon:
        icon = None
        if level in [INFO, DEBUG]:
            icon = trayicon.Information
        if level == WARNING:
            icon = trayicon.Warning
        if level in [ERROR, CRITICAL]:
            icon = trayicon.Critical
        if icon is None:
            raise ValueError("Invalid message level.")
        from .. import threads  # must be a late import

        threads.invoke_in_main_thread(trayicon.show)

        threads.invoke_in_main_thread(trayicon.showMessage, title, "".join(args), icon, timeout)
        threads.invoke_in_main_thread(lambda *_: QTimer.singleShot(timeout, trayicon.hide))
        # trayicon.showMessage(title, ''.join(args), icon, timeout)  # TODO: check if title and message are swapped?

    logMessage(*args)


def showMessage(*args, timeout=5, **kwargs):
    """
    Same as logMessage, but displays to the subscribed statusbar with a timeout.

    Parameters
    ----------
    args        :   tuple(str)
        See logMessage...
    timeout     :   int
        How long the message is displayed. If set 0, the message is persistent.
    kwargs      :   dict
        See logMessage...
    Returns
    -------

    """
    s = " ".join(args)
    if statusbar is not None:
        from .. import threads  # must be a late import

        threads.invoke_in_main_thread(statusbar.showMessage, s, timeout * 1000)

    logMessage(*args, **kwargs)


def logMessage(*args: Any, level: int = INFO, loggername: str = None, sep=" "):
    """
    Logs messages to logging log. Gui widgets can be subscribed to the log with:
        logging.getLogger().addHandler(callable)


    Parameters
    ----------
    args            : tuple[str]
        Similar to python 3's print(), any number of objects that can be cast as str. These are joined and printed as
        one line.
    level           : int
        Logging level; one of msg.DEBUG, msg.INFO, msg.WARNING, msg.ERROR, msg.CRITICAL. Default is INFO
    loggername      : str
        The name of the log to post the message into. Typically left blank, and populated by inspection.
    sep             : str
        A string to use for joining args; the separator

    """

    # Join the args to a string
    message = sep.join(map(str, args))

    if loggername is not None:
        warnings.warn("Custom loggername is no longer supported, " "ignored.")
    caller_name = sys._getframe().f_back.f_code.co_name
    logger.log(level, message, extra={"caller_name": caller_name})


def clearMessage():
    """
    Clear messages from the statusbar
    """
    from .. import threads  # must be a late import

    threads.invoke_in_main_thread(statusbar.clearMessage)


def logError(exception: Exception, value=None, tb=None, **kwargs):
    """
    Logs an exception with traceback. All uncaught exceptions get hooked here

    """

    if not value:
        value = exception
    if not tb:
        tb = exception.__traceback__
    kwargs["level"] = ERROR
    if "loggername" not in kwargs:
        kwargs["loggername"] = sys._getframe().f_back.f_code.co_name
    logMessage("\n", "The following error was handled safely by Xi-cam. It is displayed here for debugging.", **kwargs)
    try:
        logMessage("\n", *traceback.format_exception(exception, value, tb), **kwargs)
    except AttributeError:
        logMessage("\n", *traceback.format_exception_only(exception, value), **kwargs)


cumulative_time = defaultdict(lambda: 0)


@contextmanager
def logTime(*args: Any, level: int = INFO, loggername: str = None, cumulative_key: str = "", sep=" ") -> None:
    start = time.clock_gettime_ns(time.CLOCK_THREAD_CPUTIME_ID)
    yield
    elapsed_time = time.clock_gettime_ns(time.CLOCK_THREAD_CPUTIME_ID) - start

    if cumulative_key:
        cumulative_time[cumulative_key] += elapsed_time
        extra_args = [
            f"cumulative elapsed: {cumulative_time[cumulative_key] / 1e6} ms elapsed: {elapsed_time / 1e6} ms elapsed"
        ]
    else:
        extra_args = [f"elapsed: {elapsed_time / 1e6} ms elapsed"]

    logMessage(*(args + extra_args), level, loggername, sep)


@contextmanager
def busyContext() -> None:
    showBusy()
    yield
    showReady()


import sys

sys._excepthook = sys.excepthook = logError

try:
    faulthandler.enable()
except RuntimeError:
    faulthandler.enable(file=open(os.path.join(paths.user_cache_dir, "logs", "crash_log.log"), "w"))

# The above enables printing tracebacks during hard crashes. To see it in action, try the following lines
# import ctypes
# ctypes.string_at(0)



from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QPixmap, QIcon
from qtpy.QtWidgets import QLineEdit, QToolButton, QStyle

from xicam.gui.static import path


class SearchLineEdit(QLineEdit):
    def __init__(self, text="", clearable=True, parent=None):
        QLineEdit.__init__(self, text=text, parent=parent)

        searchPixmap = QPixmap(str(path("icons/search.png")))

        clearPixmap = QPixmap(str(path("icons/clear.png")))
        self.clearButton = QToolButton(self)
        self.clearButton.setIcon(QIcon(clearPixmap))
        self.clearButton.setIconSize(QSize(16, 16))
        self.clearButton.setCursor(Qt.ArrowCursor)
        self.clearButton.setStyleSheet("QToolButton { border: none; padding: 0 px;}")
        self.clearButton.hide()

        if clearable:
            self.clearButton.clicked.connect(self.clear)
            self.textChanged.connect(self.updateCloseButton)

        self.searchButton = QToolButton(self)
        self.searchButton.setIcon(QIcon(searchPixmap))
        self.searchButton.setIconSize(QSize(16, 16))
        self.searchButton.setStyleSheet("QToolButton { border: none; padding: 0 px;}")

        frameWidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.setStyleSheet(
            "QLineEdit { padding-left: %spx; padding - right: % spx;} "
            % (self.searchButton.sizeHint().width() + frameWidth + 1, self.clearButton.sizeHint().width() + frameWidth + 1)
        )
        msz = self.minimumSizeHint()
        self.setMinimumSize(
            max(msz.width(), self.searchButton.sizeHint().width() + self.clearButton.sizeHint().width() + frameWidth * 2 + 2),
            max(msz.height(), self.clearButton.sizeHint().height() + frameWidth * 2 + 2),
        )

    #        self.searchMenu = QtGui.QMenu(self.searchButton)
    #        self.searchButton.setMenu(self.searchMenu)
    #        self.searchMenu.addAction("Google")
    #        self.searchButton.setPopupMode(QtGui.QToolButton.InstantPopup)

    def resizeEvent(self, event):
        sz = self.clearButton.sizeHint()
        frameWidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.clearButton.move(self.rect().right() - frameWidth - sz.width(), (self.rect().bottom() + 1 - sz.height()) / 2)
        self.searchButton.move(self.rect().left() + 1, (self.rect().bottom() + 1 - sz.height()) / 2)

    def updateCloseButton(self, text):
        if text:
            self.clearButton.setVisible(True)
        else:
            self.clearButton.setVisible(False)



"""TODO Module docstring"""
import inspect
import weakref
from typing import Collection, Tuple, Type, Union, List, Callable, Sequence
from collections import OrderedDict
from collections.abc import MutableMapping

from pyqtgraph.parametertree import Parameter

from xicam.core import msg

from .hints import PlotHint
from .plugin import PluginType


class OperationError(Exception):
    """Base exception for this module."""
    pass


class ValidationError(OperationError):
    """Exception raised for invalid OperationPlugin configurations.

    Attributes
    ----------
    operation : OperationPlugin
        Reference to the operation that failed its validation check.
    message : str
        Explanation of the error.
    """

    def __init__(self, operation, message):
        self.operation = operation
        self.message = message

    # def __repr__(self):
    #     return f"ValidationError({self.operation!r}, {self.message!r})"

    def __str__(self):
        """Returns a readable string for this exception."""
        return f"Validation failed for {self.operation}: {self.message}"


class _OperationDict(MutableMapping):
    """Internal mapping class that has reference to an associated operation.

    Implements the MutableMapping abstract class,
    providing custom behavior for __setitem__.
    Changing the value of the internal mapping will attempt to
    update the associated operation's Parameter value.

    Context: OperationPlugin.wireup_parameter will ensure that when a Parameter's
    value or fixed state is changed, the associated operation state
    (fixed[parameter.name()], filled_value[parameter.name())is changed as well.
    This class achieves synchronization in the opposite direction -
    when the operation state changes,
    the associated Parameter value / fixed state is updated.

    For example, if a value in Operation.filled_values changes ("strength" = 0.6),
    then we need to update the associated Parameter's value (if the Parameter exists).
    """
    def __init__(self, seq=None, op=None, opts_key: str = "value", **kwargs):
        self.mapping = {}
        if seq:
            self.mapping.update(seq)
        self.mapping.update(kwargs)
        self.operation = op
        self._opts_key = opts_key

    def __delitem__(self, key):
        del self.mapping[key]

    def __getitem__(self, item):
        return self.mapping[item]

    def __setitem__(self, key, value):
        self.mapping[key] = value
        if self.operation:
            if self.operation._parameter:
                group_parameter = self.operation._parameter()
                if group_parameter:
                    try:
                        group_parameter.child(key).setOpts(**{self._opts_key: value})
                        # group_parameter.child(key).setValue(value,
                        #                                     blockSignal=self._operation._set_value)
                    except Exception as e:
                        msg.logError(e)

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f"{type(self).__name__}({self.mapping}, op={self.operation})"

    def copy(self):
        return _OperationDict(self.mapping.copy(), op=self.operation, opts_key=self._opts_key)


class OperationPlugin(PluginType):
    """A plugin that can be used to define an operation, which can be used in a Workflow.

    Note: use the `@operation` decorator to create an operation type.

    At its simplest level, an operation can be though of as a function.
    Any arguments (parameters) defined in the python function are treated as inputs for the operation.
    An operation's outputs are defined by the returned values of the python function.

    There are various methods available to help with modifying the operation's parameters.

    For more information on the attributes, see the documentation for their respective method
    (e.g. for more information on `limits`, see the `limits` method documentation).

    For an easy way to expose parameters in the GUI, use `OperationPlugin.as_parameter` in conjunction with
    `pyqtgraph.Parameter.create`.
    Note that only input parameters that have type hinting annotations will be included in the return value
    of `OperationPlugin.as_parameter`.

    Attributes
    ----------
    filled_values : dict
        Keys are the parameter names, values are the current values for the parameter.
    fixable : dict
        Keys are the parameter names, values are bools indicating whether or not the parameter
        is able to be fixed.
    fixed : dict
        Keys are the parameter names, values are bools indicating whether or not the parameter
        is fixed.
    input_names : Tuple[str, ...]
        Names (in order) of the input argument(s) for the operation. Note that if not provided,
        input names default to the argument names in the function signature.
    limits : dict
        Keys are the parameter names, values are the limits (which are a collection of floats).
    opts : dict
        Any additional options (kwargs) to be passed to the parameter (useful with pyqtgraph).
    output_names : Tuple[str, ...]
        Names (in order) of the output(s) for the operation.
    output_shape : dict
        Keys are the output parameter names, values are the expected shape of the output
        (which are of type list).
    units : dict
        Keys are the parameter names, values are units (of type str).
    visible : dict
        Keys are the pareameter names, values are bools indicating whehter or not the parameter
        is visible (when exposed using pyqtgraph).
    disabled : bool
        Whether or not the operation is disabled (default is False).
    display_name: str
        The name of the operation as it should be displayed to a user.
    hints : list
    input_descriptions : dict
        A mapping dict containing descriptions of each named input parameter
    output_descriptions : dict
        A mapping dict containing descriptions of each named output parameter

    Notes
    -----
    This class formally deprecates usage of the ProcessingPlugin API.

    Example
    --------
    Here, we define a function, then wrap it with the OperationPlugin decorator to make it an operation.

    >>> @OperationPlugin
    >>> def my_operation(x: int = 1, y: int = 2): -> int
    >>>     return x + y

    """
    needs_qt = False

    _func = None  # type: Callable
    filled_values = _OperationDict()  # type: _OperationDict
    fixable = {}  # type: dict
    fixed = _OperationDict(opts_key="fixed") # type: _OperationDict
    input_names = None  # type: Tuple[str]
    output_names = None  # type: Tuple[str]
    limits = {}  # type: dict
    opts = {}  # type: dict
    output_shape = {}  # type: dict
    units = {}  # type: dict
    visible = {}  # type: dict
    name = None  # type: str
    input_descriptions = {}  # type: dict
    output_descriptions = {}  # type: dict
    categories = None  # type: Sequence[Union[tuple, str]]
    hints = []

    def __init__(self, **filled_values):
        super(OperationPlugin, self).__init__()
        # Copy class dict information so that changes to instance don't propagate to class
        self.filled_values = self.filled_values.copy()
        self.filled_values.operation = self  # Need to add the operation ref to our OperationDict
        self.filled_values.update(filled_values)
        self.fixable = self.fixable.copy()
        self.fixed = self.fixed.copy()
        self.fixed.operation = self  # Need to add the operation ref to our OperationDict
        self.limits = self.limits.copy()
        self.opts = self.opts.copy()
        self.output_shape = self.output_shape.copy()
        self.units = self.units.copy()
        self.hints = self.hints.copy()
        self._parameter = None  # type: weakref.ref

    @classmethod
    def _validate(cls):
        """Validates the OperationPlugin's inputs and outputs."""

        # Capture any validation issues for use later when raising
        invalid_msg = ""
        # Define which "input" arg properties we want to check
        input_properties = {"fixable": cls.fixable,
                            "fixed": cls.fixed,
                            "limits": cls.limits,
                            "opts": cls.opts,
                            "units": cls.units,
                            "visible": cls.visible}
        # Check if all the attributes have a default value for each input param
        for arg in cls.input_names:
            for name, prop in input_properties.items():
                if prop and arg not in prop:
                    pass  # Do we want to enforce input default values?

        # Check if there is a 1:1 mapping from user-specified input_names to function args
        num_names = len(cls.input_names)
        num_args = len(inspect.signature(cls._func).parameters.keys())
        if num_names != num_args:
            invalid_msg += (f"Number of input_names given ({num_names}) "
                            f"must match number of inputs for the operation ({num_args}).")
        # Check if there are any input args that are not actually defined in the operation
        # e.g. 'x' is not a valid input in the case below:
        # @visible('x')
        # def func(a): return
        for name, prop in input_properties.items():
            for arg in prop.keys():
                if arg not in cls.input_names:
                    invalid_msg += f"\"{arg}\" is not a valid input for \"{name}\". "

        # Warn if there are no output_names defined
        if not len(cls.output_names):
            warning_msg = (f"No output_names have been specified for your operation {cls}; "
                           f"you will not be able to connect your operation's output(s) to "
                           f"any other operations.")
            msg.logMessage(warning_msg, level=msg.WARNING)

        # Define which "output" arg properties we want to check
        output_properties = {"output_shape": cls.output_shape}
        # Check if there are any output args that are not actually defined in the operation
        for name, prop in output_properties.items():
            for arg in prop.keys():
                if arg not in cls.output_names:
                    invalid_msg += f"\"{arg}\" is not a valid output for \"{name}\". "

        if invalid_msg:
            raise ValidationError(cls, invalid_msg)
        else:
            msg.logMessage(f"All args for {cls} are valid.")

    def __call__(self, **kwargs):
        """Allows this class to be used as a function decorator."""
        filled_kwargs = self.filled_values.copy()
        filled_kwargs.update(kwargs)
        return self._func(**filled_kwargs)

    def __str__(self):
        return f"OperationPlugin named {self.name}"

    @property
    def input_types(self) -> 'OrderedDict[str, Type]':
        """Returns the types of the inputs for the operation."""
        signature = inspect.signature(self._func)
        input_type_map = OrderedDict([(name, parameter.annotation) for name, parameter in signature.parameters.items()])
        return input_type_map

    @property
    def output_types(self) -> 'OrderedDict[str, Type]':
        """Returns the types of the outputs for the operation."""
        return_annotation = inspect.signature(self._func).return_annotation
        if not return_annotation or return_annotation is inspect.Signature.empty:
            return_annotation = tuple()

        output_type_map = OrderedDict(zip(self.output_names, return_annotation.__args__))
        return output_type_map

    def __reduce__(self):
        return OperationPlugin, tuple(), {'_func': self._func,
                                          'filled_values': self.filled_values,
                                          'input_names': self.input_names,
                                          'output_names': self.output_names}

    def as_parameter(self):
        """Return the operation's inputs as a ready-to-use object with pyqtgraph.

        A list of dictionaries is returned with each dictionary representing one of the operation's input parameters.
        Each dictionary represents the state of the input parameter;
        for example, its name, its default value, its type, etc.
        Note that only inputs that have been annotated with type-hinting
        and whose types are registered with pyqtgraph (PARAM_TYPES) will be included in this list.
        This list can be passed to `pyqtgraph.Parameter.create` to create a parameter tree widget.

        Alternative text:
        A list of dictionaries is returned where each dict is a best-effort attempt to represent each input parameter as a pyqtgraph Parameter.

        Returns
        -------
        parameters : list
            List of dictionaries; each dictionary represents the state of an input parameter
            (only applies to input parameters that are annotated with type-hinting).

        Notes
        -----
        For more information about pyqtgraph and Parameters,
        see http://www.pyqtgraph.org/documentation/parametertree/parameter.html?highlight=create#pyqtgraph.parametertree.Parameter.create
        """
        from pyqtgraph.parametertree.Parameter import PARAM_TYPES

        parameter_dicts = []
        for name, parameter in inspect.signature(self._func).parameters.items():
            if getattr(parameter.annotation, '__name__', None) in PARAM_TYPES:
                parameter_dict = dict()
                parameter_dict.update(self.opts.get(name, {}))
                parameter_dict['name'] = name
                parameter_dict[
                    'default'] = parameter.default if parameter.default is not inspect.Parameter.empty else None
                parameter_dict['value'] = self.filled_values[
                    name] if name in self.filled_values else parameter_dict['default']

                parameter_dict['type'] = getattr(self.input_types[name], '__name__', None)
                if name in self.limits:
                    parameter_dict['limits'] = self.limits[name]
                parameter_dict['units'] = self.units.get(name)
                parameter_dict['fixed'] = self.fixed.get(name)
                parameter_dict['fixable'] = self.fixable.get(name)
                parameter_dict['visible'] = self.visible.get(name, True)
                parameter_dict.update(self.opts.get(name, {}))

                parameter_dicts.append(parameter_dict)

            elif getattr(self.input_types[name], "__name__", None) == "Enum":
                parameter_dict = dict()
                parameter_dict['name'] = name
                parameter_dict['value'] = self.filled_values[
                    name] if name in self.filled_values else parameter.default
                parameter_dict['values'] = self.limits.get(name) or ["---"],
                parameter_dict['default'] = parameter.default
                parameter_dict['type'] = "list",
                if name in self.limits:
                    parameter_dict['limits'] = self.limits[name]
                parameter_dict['units'] = self.units.get(name)
                parameter_dict['fixed'] = self.fixed.get(name)  # TODO: Does this need a default value
                parameter_dict['fixable'] = self.fixable.get(name)
                parameter_dict['visible'] = self.visible.get(name, True)  # TODO: should we store the defaults at top?
                parameter_dict.update(self.opts.get(name, {}))

                parameter_dicts.append(parameter_dict)
        return parameter_dicts

    def wireup_parameter(self, parameter: Parameter):
        """Wire up a Parameter (created from an operation) to update the operation's state.

        Updates the operation's values (filled_values) if the Parameter values are changed.
        Updates the operatin's fixed state if the Parameter's fixed state is changed.

        This is useful when you want to interact with an operations inputs in a GUI
        by using something like a ParameterTree (see example).

        Parameters
        ----------
        parameter : Parameter
            A pyqtgraph Parameter created from an operation (via `as_parameter()`)

        Example
        -------
        >>> from pyqtgraph.parametertree import ParameterTree
        >>> from pyqtgraph.parametertree.parameterTypes import GroupParameter
        >>> from qtpy.QtWidgets import QApplication, QMainWindow
        >>> from xicam.plugins.operationplugin import operation, output_names
        >>> @operation
        >>> @output_names("sum")
        >>> def my_operation(x: float = 1.0, y: float = 2.0):
        >>>     return x + y
        >>> qapp = QApplication([])
        >>> parameter_tree = ParameterTree()
        >>> op = my_operation()
        >>> parameter = GroupParameter(name=op.name, children=op.as_parameter())
        >>> op.wireup_parameter(parameter)
        >>> parameter_tree.addParameters(parameter)
        >>> window = QMainWindow()
        >>> window.setCentralWidget(parameter_tree)
        >>> window.show()
        >>> qapp.exec_()
        """
        for child, param in zip(parameter.children(), self.as_parameter()):
            # wireup signals to update the workflow
            if param.get('fixable'):
                child.sigFixToggled.connect(self._set_fixed)
            msg.logMessage(child.name())
            # child.sigValueChanged.connect(lambda *args: print(args))
            child.sigValueChanged.connect(self._set_value)
            # We want to propagate changes to the operation's fixed/values state to the Parameter
            self._parameter = weakref.ref(parameter)

    def _set_fixed(self, param: Parameter, value):
        """Update the fixed state for the operation when it is toggled in the corresponding Parameter."""
        self.fixed[param.name()] = value

    def _set_value(self, param: Parameter, value):
        """Update the value for an operation when it is changed in he corresponding Parameter."""
        self.filled_values[param.name()] = value


def operation(func: Callable,
              filled_values: dict = None, fixable: dict = None, fixed: dict = None,
              input_names: Tuple[str, ...] = None, output_names: Tuple[str, ...] = None,
              limits: dict = None,
              opts: dict = None,
              output_shape: dict = None,
              units: dict = None,
              visible: dict = None,
              name: str = None,
              input_descriptions: dict = None, output_descriptions: dict = None,
              categories: Sequence[Union[tuple, str]] = None) -> Type[OperationPlugin]:
    """Create a new operation.

    *When you define a new operation, you must use this decorator (`@operation`)
    and the `@output_names` decorator.*

    This function can be used as a decorator to define a new operation type.
    The operation can then be instantiated by using the `()` operator on
    the operation function's name.

    Parameters
    ----------
    func : Callable
        Function that this operation will call.
    filled_values : dict, optional
        Values to fill for the parameters.
    fixable : dict, optional
        Indicates which parameters are able to be fixed.
    fixed :  dict, optional
        Indicates whether or not a parameter is fixed.
    limits : dict, optional
        Defines limits for parameters.
    opts : dict, optional
        Additional options (kwargs) for the parameter
        (useful with pyqtgraph's Parameter/ParameterTree).
    output_names : tuple, optional
        Names for the outputs, or returned values, of the operation.
    output_shape : dict, optional
        Defines expected shapes for the outputs.
    units : dict, optional
        Defines units for the parameters in the operation.
    name : str, optional
        The display name to be shown to the user. Defaults to self.__name__
    visible : dict, optional
        Indicates if a parameter is visible or not (see pyqtgraph.Parameter).
    input_descriptions : dict, optional
        A mapping dict containing descriptions for each named input
    output_descriptions : dict, optional
        A mapping dict containing descriptions for each named output
    categories : List[Union[tuple, str], optional
        A sequence of categories to associate with this operation.

    Example
    --------
    Create a new operation type and create a new operation instance from it.

    >>> from xicam.core.execution import Workflow
    >>> from xicam.plugins.operationplugin import operation, output_names
    >>> @operation
    >>> @output_names("my_output")
    >>> def my_func(x: float = 0.0) -> float:
    >>>     return x * -1
    >>> op = my_func()
    >>> workflow = Workflow()
    >>> result = workflow.execute(x=2.5).result()
    >>> print(result)

    """
    # For now, any C functions that don't have the required metadata for inspection will raise an exception
    # TODO: link to documentation
    try:
        inspect.signature(func)
    except ValueError as e:
        msg = f"The provided function '{func.__name__}' could not be inspected." \
               "In the case that your function is a C function, you will need to wrap it. " \
               "For an example, see the Xi-CAM.SAXS.processing.inpaint module."
        raise OperationError(msg) from e

    # Allow passing a string
    if type(input_names) is str:
        input_names = (input_names,)

    if type(output_names) is str:
        output_names = (output_names,)

    state = {  # "_func": func,
        "name": name or getattr(func, 'name', getattr(func, '__name__', None)),
        # Fallback to inspecting the function arg names if no input names provided
        "input_names": input_names or getattr(func,
                                              'input_names',
                                              tuple(inspect.signature(func).parameters.keys())),
        "output_names": output_names or getattr(func, 'output_names', getattr(func, "__name__", tuple())),
        "output_shape": output_shape or getattr(func, 'output_shape', {}),
        "input_description": input_descriptions or getattr(func, 'input_descriptions', {}),
        "output_descriptions": output_descriptions or getattr(func, 'output_descriptions', {}),
        "categories": categories or getattr(func, 'categories', []),
        "filled_values": _OperationDict(filled_values) or _OperationDict(),
        "limits": limits or getattr(func, 'limits', {}),
        "units": units or getattr(func, 'units', {}),
        "fixed": _OperationDict(fixed, opts_key="fixed") or
                 _OperationDict(getattr(func, 'fixed', {}), opts_key="fixed"),
        "fixable": fixable or getattr(func, 'fixable', {}),
        "visible": visible or getattr(func, 'visible', {}),
        "opts": opts or getattr(func, 'opts', {}),
        "hints": getattr(func, 'hints', [])  # TODO: does hints need an arg
    }

    if state["name"] is None:
        raise NameError('The provided operation is unnamed.')

    if type(state["output_names"]) == str:
        state["output_names"] = (state["output_names"],)

    operation_class = type("WrappedOperationPlugin", (OperationPlugin,), state)  # Ignore intellisense warnings
    operation_class._func = staticmethod(func)

    operation_class._validate()

    return operation_class


def _quick_set(func, attr_name, key, value, init):
    # TODO: does this need to be called initially to provide valid defaults?
    if not hasattr(func, attr_name):
        setattr(func, attr_name, init)
    getattr(func, attr_name)[key] = value


def display_name(name):
    """Set the display name for the operation.

    Display name is how this operation's name will be displayed in Xi-cam.

    Parameters
    ----------
    name : str
        Name for the operation.

    Example
    --------
    Create an operation whose display name is "Cube Operation."

    >>> @OperationPlugin
    >>> @display_name('Cube Operation')
    >>> def cube(n: int = 2) -> int:
    >>>     return n**3
    """

    def decorator(func):
        func.name = name
        return func

    return decorator


def units(arg_name, unit):
    """Decorator to define units for an input.

    Associates a unit of measurement with an input.

    Parameters
    ----------
    arg_name : str
        Name of the input to attach a unit to.
    unit : str
        Unit of measurement descriptor to use (e.g. "mm").

    Example
    --------
    Create an operation where its `x` parameter has its units defined in microns.

    >>> @OperationPlugin
    >>> @units('x', '\u03BC'+'m')
    >>> def op(x: float = -1) -> float:
    >>>     return x *= -1.0
    """

    def decorator(func):
        _quick_set(func, 'units', arg_name, unit, {})
        return func

    return decorator


def fixed(arg_name, fix=True):
    """Decorator to set whether or not an input's value is fixed.

    Fixed means that the input's value is fixed in the context of model fitting.

    By default, sets the `arg_name` input to fixed,
    meaning its value cannot be changed.

    Parameters
    ----------
    arg_name : str
        Name of the input to change fix-state for.
    fix : bool, optional
        Whether or not to fix `arg_name` (default is True).

    TODO example
    """

    def decorator(func):
        _quick_set(func, 'fixed', arg_name, fix, {})
        # TODO do we need to make 'fixed' args 'fixable'/'readonly'???
        # _quick_set(func, 'fixable', arg_name, True, {})
        return func

    return decorator


def limits(arg_name, limit):
    """Decorator to define limits for an input.

    Limits restrict the allowable values for the input
    (inclusive lower-bound, inclusive upper-bound).

    Parameters
    ----------
    arg_name : str
        Name of the input to define limits for.
    limit : tuple[float]
        A 2-element sequence representing the lower and upper limit.

    Example
    --------
    Make an operation that has a limit on the `x` parameter from [0, 100].

    >>> @OperationPlugin
    >>> @limits('x', [0, 100])
    >>> def op(x):
    >>>     ...

    Make an operation that has a limit on the `x` parameter from [0.0, 1.0].

    >>> @OperationPlugin
    >>> @limits('x', [0.0, 1.0])
    >>> @opts('x', step=0.1)
    >>> def op(x):
    >>>     ...

    """

    def decorator(func):
        _quick_set(func, 'limits', arg_name, limit, {})
        return func

    return decorator


# TODO: need an image_hint decorator? coplot_hint decorator?

# TODO Check that signature propagates up
def plot_hint(*args, **kwargs):
    """Decorator to define plot hints for 1-dimensional outputs. This accepts `kwargs` that get passed to the rendering
    widget; see :class:`pyqtgraph.PlotDataItem`. This annotation "hints" to Xi-cam that a Plot widget may be generated
    to represent this data visually.

    >>> @OperationPlugin
    >>> @output_names('x', 'y')
    >>> @plot_hint('x', 'y', yLog=True, labels={"left": "y-axis", "bottom": "x-axis"}, pen='r')
    >>> def op(x):
    >>>     ...

    Parameters
    ----------
    args
        Arguments for `PlotHint`.
    kwargs
        Keyword arguments for `PlotHint`.

    TODO examples may be helpful in these...
    """

    def decorator(func):
        if not hasattr(func, 'hints'):
            func.hints = []
        func.hints.append(PlotHint(*args, **kwargs))
        return func

    return decorator


def input_names(*names):
    """Decorator to define input names for the operation.

    The number of names provided must match the number of arguments for the operation/function.

    If not provided, input names will be determined by examining the names of the arguments
    to the operation function.

    Example
    --------
    Create an addition operation and use the names "first" and "second" for the input names
    instead of the function arg names (x and y).

    >>> @OperationPlugin
    >>> @input_names("first", "second")
    >>> def my_add(x: int, y: int) -> int:
    >>>     return x + y
    """

    def decorator(func):
        func.input_names = names
        return func

    return decorator


def output_names(*names):
    """Decorator to define the names of the outputs for an operation.

    Defines N-number of output names. These names will be used (in-order)
    to define any outputs that the operation has.

    Parameters
    ----------
    names : List[str]
        Names for the outputs in the operation.

    Example
    --------
    Define an operation that has the outputs `x` and `y`.

    >>> @OperationPlugin
    >>> @output_names("x", "y")
    >>> def some_operation(a: int, b: int) -> Tuple[int, int]:
    >>>     return a, b

    """

    def decorator(func):
        func.output_names = names
        return func

    return decorator


def output_shape(arg_name: str, shape: Union[int, Collection[int]]):
    # TODO: how does this work? How do we know the shape before runtime?
    """Decorator to set the shape of an output in an operation."

    Parameters
    ----------
    arg_name : str
        Name of the output to define a shape for.
    shape : int or tuple of ints
        N-element tuple representing the shape (dimensions) of the output.

    Example
    --------
    TODO
    """

    def decorator(func):
        _quick_set(func, 'output_shape', arg_name, shape, {})
        return func

    return decorator


def visible(arg_name: str, is_visible=True):
    """Decorator to set whether an input is visible (shown in GUI) or not.

    Parameters
    ----------
    arg_name : str
        Name of the input to change visibility for.
    is_visible : bool, optional
        Whether or not to make the input visible or not (default is True).

    Example
    --------
    Define an operation that makes the data_image invisible to the GUI (when using `as_parameter()` and pyqtgraph).

    >>> @OperationPlugin
    >>> @visible('data_image')
    >>> def threshold(data_image: np.ndarray, threshold: float = 0.5) -> np.ndarray:
    >>>     return ...

    """

    def decorator(func):
        _quick_set(func, 'visible', arg_name, is_visible, {})
        return func

    return decorator


def opts(arg_name: str, **options):
    """Decorator to set the opts (pyqtgraph Parameter opts) for `arg_name`.

    This is useful for attaching any extra attributes onto an operation input argument.

    These options correspond to the optional opts expected by pyqtgraph.Parameter.
    The options are typically used to add extra configuration to a Parameter.

    Parameters
    ----------
    arg_name : str
        Name of the input to add options for.
    options : keyword args
        Keyword arguments that can be used for the rendering backend (pyqtgraph).

    Example
    --------
    Define an operation where the `x` input is readonly.

    >>> @OperationPlugin
    >>> @opts('x', 'readonly'=True)
    >>> def op(x: str = 100) -> str:
    >>>     return x
    """

    def decorator(func):
        _quick_set(func, 'opts', arg_name, options, {})
        return func

    return decorator


def _describe_arg(arg_type: str, arg_name: str, description: str):
    assert arg_type in ['input', 'output']

    def decorator(func):
        _quick_set(func, f'{arg_type}_descriptions', arg_name, description, {})
        return func

    return decorator


def describe_input(arg_name: str, description: str):
    """Decorator to set the description for input `arg_name`.

    This is useful for annotating the parameter with additional information for users.

    These annotations are displayed in GUI representations of the operation.

    Parameters
    ----------
    arg_name : str
        Name of the input to add options for.
    description : str
        A human-readable description of the input `arg_name`

    Example
    --------
    Define an operation and attach a description to its `x` input argument.

    >>> @OperationPlugin
    >>> @describe_input('x', 'The value to square.')
    >>> def square(x: int = 100) -> int:
    >>>     return x**2
    """

    return _describe_arg('input', arg_name, description)


def describe_output(arg_name: str, description: str):
    """Decorator to set the description for output `arg_name`.

    This is useful for annotating the parameter with additional information for users.

    These annotations are displayed in GUI representations of the operation.

    Parameters
    ----------
    arg_name : str
        Name of the input to add options for.
    description : str
        A human-readable description of the output `arg_name`.

    Example
    --------
    Define an operation and attach a description to its `square` output.

    >>> @OperationPlugin
    >>> @output_names('square')
    >>> @describe_output('square', 'The squared value of x.')
    >>> def square(x: int = 100) -> int:
    >>>     return x**2
    """

    return _describe_arg('output', arg_name, description)


def categories(*categories: Tuple[Union[tuple, str]]):
    """Decorator to assign categories to a operation.

    These categories will be used to populate the structure of Xi-cam's menus of `OperationPlugins`.

    Parameters
    ----------
    categories : Tuple[Union[tuple, str]]
        A sequence of categories. Each item is a tuple or str. If an item is a tuple, each item in the tuple is considered
        as an additional depth in the menu structure.

    Example
    --------
    Define an operation that is in the following categories::

        Generic Functions
            Simple Math
                (square operation)
        Math Functions
                (square operation)

    >>> @OperationPlugin
    >>> @categories(('Generic Functions', 'Simple Math'), 'Math Functions')
    >>> def square(x: int = 100) -> int:
    >>>     return x**2
    """

    def decorator(func):
        if not hasattr(func, 'categories'):
            func.categories = []
        func.categories.append(categories)
        return func

    return decorator



import pytest

import numpy as np

from xicam.core import msg

# from xicam.plugins import operation
from xicam.plugins.operationplugin import (
    display_name,
    fixed,
    input_names,
    limits,
    opts,
    output_names,
    output_shape,
    plot_hint,
    units,
    visible,
    ValidationError,
    operation,
)


# Tests both the function interface and Operation API interface
#
# the "test_defaults" cases test the absence of the tested decorator,
# i.e. what is the default initialized value for the tested decorator's
# associated attribute

# @fixed
class TestFixed:
    def test_defaults(self):
        def func(a, b):
            return

        # assert func.fixed == {}
        assert not hasattr(func, "fixed")
        # Test Operation API
        op = operation(func)
        assert op.fixed == {}

    def test_single(self):
        @fixed("a")
        def func(a, b):
            return

        assert func.fixed == {"a": True}
        # Test Operation API
        op = operation(func)
        assert op.fixed == {"a": True}

    def test_multiple(self):
        @fixed("a")
        @fixed("b")
        def func(a, b, c):
            return

        assert func.fixed == {"a": True, "b": True}
        # Test Operation API
        op = operation(func)
        assert op.fixed == {"a": True, "b": True}

    def test_explicit(self):
        @fixed("a", True)
        @fixed("b", False)
        def func(a, b, c):
            return

        assert func.fixed == {"a": True, "b": False}
        # Test Operation API
        op = operation(func)
        assert op.fixed == {"a": True, "b": False}

    def test_redundant(self):
        @fixed("b")
        @fixed("b")
        def func(a, b, c):
            return

        assert func.fixed == {"b": True}
        # Test Operation API
        op = operation(func)
        assert op.fixed == {"b": True}

    def test_no_parameter_with_name(self):
        @fixed("dne")
        def func(a, b, c):
            return

        assert func.fixed == {"dne": True}
        # Test Operation API
        with pytest.raises(ValidationError):
            operation(func)

    def test_bad(self):
        # TODO: do we need to test unexpected types?
        @fixed(3)
        def func(a):
            return

        assert True


# @limits
class TestLimits:
    def test_defaults(self):
        def func(a):
            return

        assert not hasattr(func, "limits")
        # Test Operation API
        op = operation(func)
        assert op.limits == {}

    def test_limits(self):
        @limits("a", [0.0, 1.0])
        def func(a):
            return

        assert func.limits == {"a": [0.0, 1.0]}
        # Test Operation API
        op = operation(func)
        assert op.limits == {"a": [0.0, 1.0]}


# @output_names
class TestOutputNames:
    def test_output_names(self):
        @output_names("sum")
        def my_sum(a, b):
            return np.sum(a, b)

        assert my_sum.output_names == ("sum",)
        # Test Operation API
        op = operation(my_sum)
        assert op.output_names == ("sum",)

    def test_output_names_none_provided(self, caplog):
        def my_op(a, b):
            return 42

        assert hasattr(my_op, "output_names") is False
        # Test Operation API
        op = operation(my_op)
        assert op.output_names == ("my_op", )
        # Expecting a msg.WARNING log record that mentions "output_names"
        expected_warn_record = caplog.records[0]
        assert expected_warn_record.levelno == msg.WARNING
        assert "output_names" in expected_warn_record.msg


# @input_names
class TestInputNames:
    def test_defaults(self):
        # Test not using the input_names decorator
        def my_op(x, y):
            return x + y

        assert hasattr(my_op, "input_names") is False
        # Test Operation API
        op = operation(my_op)
        assert op.input_names == ("x", "y")

    def test_matching_number_of_names(self):
        # Test a good use of the input_names decorator
        @input_names("first", "second")
        def my_sum(x, y):
            return x + y

        assert my_sum.input_names == ("first", "second")
        # Test the Operation API
        op = operation(my_sum)
        assert op.input_names == ("first", "second")

    def test_fewer_input_names(self):
        # Test when there are fewer input names given than there are function args
        @input_names("first")
        def my_sum(x, y):
            return x + y

        assert my_sum.input_names == ("first",)
        # Test the Operation API
        with pytest.raises(ValidationError):
            operation(my_sum)

    def test_extra_input_names(self):
        # Test when there are more input names given that there are function args
        @input_names("first", "second", "third", "fourth")
        def my_sum(x, y):
            return x + y

        assert my_sum.input_names == ("first", "second", "third", "fourth")
        # Test the Operation API
        with pytest.raises(ValidationError):
            operation(my_sum)


class TestOutputShape:
    def test_output_shape(self):
        # TODO: what should the expected shape be? A collection (list/tuple)?
        @output_shape("out", (10, 10))
        @output_names("out")
        def func(a) -> np.ndarray:
            return np.zeros(shape=(10, 10))

        assert func.output_shape == {"out": (10, 10)}
        # Test Operation API
        op = operation(func)
        assert op.output_shape == {"out": (10, 10)}

    def test_missing_output_name(self):
        # TODO: what should the expected shape be? A collection (list/tuple)?
        @output_shape("out", (10, 10))
        def func(a) -> np.ndarray:
            return np.zeros(shape=(10, 10))

        assert func.output_shape == {"out": (10, 10)}
        # Test Operation API
        with pytest.raises(ValidationError):
            operation(func)


# def test_plot_hint():
#     print('tlaktj')
#     assert False, 'bnlah'


# @units
class TestUnits:
    def test_defaults(self):
        def func(a):
            return

        assert not hasattr(func, "units")
        # Test Operation API
        op = operation(func)
        assert op.units == {}

    def test_units(self):
        @units("a", "mm")
        def func(a):
            return

        assert func.units == {"a": "mm"}
        # Test Operation API
        op = operation(func)
        assert op.units == {"a": "mm"}


# @display_name
class TestDisplayName:
    def test_defaults(self):
        def func(a):
            return

        assert not hasattr(func, "name")
        # Test Operation API
        op = operation(func)
        assert op.name == func.__name__

    def test_display_name(self):
        @display_name("my operation name")
        def func(a):
            return

        assert func.name == "my operation name"
        # Test Operation API
        op = operation(func)
        assert op.name == "my operation name"


# @visible
class TestVisible:
    def test_defaults(self):
        # TODO: what should this return?
        @output_names("z")
        def func(a):
            return

        # assert func.visible == {'a': True}  # Expects a default to be set regardless
        assert not hasattr(func, "visible")  # does not expect any defaults to be set; must be explicit
        # Test Operations API
        op = operation(func)
        assert op.visible == {}

    def test_default(self):
        @visible("a")
        def func(a, b):
            return

        assert func.visible == {"a": True}
        # Test Operations API
        op = operation(func)
        assert op.visible == {"a": True}

    def test_true(self):
        @visible("a", True)
        def func(a, b):
            return

        assert func.visible == {"a": True}
        # Test Operations API
        op = operation(func)
        assert op.visible == {"a": True}

    def test_false(self):
        @visible("a", False)
        def func(a, b):
            return

        assert func.visible == {"a": False}
        # Test Operations API
        op = operation(func)
        assert op.visible == {"a": False}


# @opts
class TestOpts:
    def test_defaults(self):
        def func(a):
            return

        assert not hasattr(func, "opts")
        # Test Operation API
        op = operation(func)
        assert op.opts == {}

    def test_opts(self):
        # TODO: what happens if you pass in an 'invalid' opt? what is an invalid opt?
        @opts("a", someopt="opt")
        def func(a, b):
            return

        assert func.opts == {"a": {"someopt": "opt"}}
        # Test Operation API
        op = operation(func)
        assert op.opts == {"a": {"someopt": "opt"}}


def test_as_parameter():
    @operation
    # @display_name('my_operation_name')
    @fixed("fixed_param")
    # @input_names('fixed', 'visible', 'limits', 'opts', 'units', 'unseen', 'default')
    @limits("limits_param", [0, 100])
    @opts("opts_param", opt="value")
    @output_names("out1", "out2")
    @output_shape("out1", [1])
    @units("units_param", "km")
    @visible("invisible_param", False)
    def func(
        fixed_param: int,
        invisible_param: int,
        limits_param: int,
        opts_param: list,
        units_param: float,
        unseen_param,
        default_param: int = 0,
    ):
        return

    expected_as_parameter = [
        {
            "default": None,
            "fixable": None,
            "fixed": True,
            "name": "fixed_param",
            "type": "int",
            "units": None,
            "value": None,
            "visible": True,
        },
        {
            "default": None,
            "fixable": None,
            "fixed": None,
            "name": "invisible_param",
            "type": "int",
            "units": None,
            "value": None,
            "visible": False,
        },
        {
            "default": None,
            "fixable": None,
            "fixed": None,
            "limits": [0, 100],
            "name": "limits_param",
            "type": "int",
            "units": None,
            "value": None,
            "visible": True,
        },
        {
            "default": None,
            "fixable": None,
            "fixed": None,
            "name": "opts_param",
            "opt": "value",
            "type": "list",
            "units": None,
            "value": None,
            "visible": True,
        },
        {
            "default": None,
            "fixable": None,
            "fixed": None,
            "name": "units_param",
            "type": "float",
            "units": "km",
            "value": None,
            "visible": True,
        },
        {
            "default": 0,
            "fixable": None,
            "fixed": None,
            "name": "default_param",
            "type": "int",
            "units": None,
            "value": 0,
            "visible": True,
        },
    ]

    assert func().as_parameter() == expected_as_parameter


# TODO: BrokenPipe and ValueError: I/O operation on closed file exceptions occur:
# * more than one of these workflow tests is run
# * one of the workflow tests is run, and a test above has a print() in it


def test_workflow():
    from xicam.core.execution.workflow import Workflow
    from xicam.core.execution.daskexecutor import DaskExecutor
    from xicam.plugins.operationplugin import output_names

    executor = DaskExecutor()

    @operation
    @output_names("square")
    def square(a=3) -> int:
        return a ** 2

    @operation
    @output_names("sum")
    def my_sum(a, b=3) -> int:
        return a + b

    wf = Workflow()

    square = square()
    my_sum = my_sum()

    wf.add_operation(square)
    wf.add_operation(my_sum)
    wf.add_link(square, my_sum, "square", "a")

    assert wf.execute_synchronous(executor=executor) == [{"sum": 12}]


@pytest.fixture
def square_op():
    @operation
    @output_names("square")
    def square(a=3) -> int:
        return a ** 2

    return square


@pytest.fixture
def sum_op():
    @operation
    @output_names("sum")
    def my_sum(a, b=3) -> int:
        return a + b

    return my_sum


def test_multiple_instances(square_op, sum_op):
    from xicam.core.execution.workflow import Workflow
    from xicam.core.execution.daskexecutor import DaskExecutor

    executor = DaskExecutor()

    wf = Workflow()

    square = square_op()
    square2 = square_op()
    square2.filled_values["a"] = 2
    my_sum = sum_op()

    wf.add_operation(square)
    wf.add_operation(square2)
    wf.add_operation(my_sum)
    wf.add_link(square, my_sum, "square", "a")
    wf.add_link(square2, my_sum, "square", "b")

    assert wf.execute_synchronous(executor=executor) == [{"sum": 13}]


#
#
# def test_autoconnect():
#     from xicam.core.execution.workflow import Workflow
#     from xicam.core.execution.daskexecutor import DaskExecutor
#     from xicam.plugins.operation import output_names
#
#     executor = DaskExecutor()
#
#     @operation
#     @output_names('square')
#     def square(a=3) -> int:
#         return a ** 2
#
#     @operation
#     @output_names('sum')
#     def my_sum(square, b=3) -> int:
#         return square + b
#
#     wf = Workflow()
#
#     wf.add_operation(square)
#     wf.add_operation(my_sum)
#     wf.auto_connect_all()
#
#     assert wf.execute_synchronous(executor=executor) == [{'sum': 12}]



from databroker.core import BlueskyRun


def ndims_from_descriptor(descriptor: dict, field: str):
    return len(descriptor["data_keys"][field]["shape"])  # NOTE: this doesn't include event dim


def shape_from_descriptor(descriptor: dict, field: str):
    return descriptor["data_keys"][field]["shape"]


def fields_from_stream(run: BlueskyRun, stream: str):
    return fields_from_descriptor(descriptors_from_stream(run, stream))


def descriptors_from_stream(run: BlueskyRun, stream: str):
    return run[stream].metadata["descriptors"]


def fields_from_descriptor(descriptor):
    return list(descriptor["data_keys"].keys())


def streams_from_run(run: BlueskyRun):
    return list(run)


def xarray_from_run(run: BlueskyRun, stream: str = None, field: str = None):
    data = run.to_dask()

    if stream:
        data = data[stream]

        if field:
            data = data[field]

    return data



"""
Catalog Plugin for browsing pre-configured databroker catalogs
"""
from xicam.plugins.catalogplugin import CatalogPlugin
from xicam.gui.widgets.dataresourcebrowser import QListView
from databroker.core import BlueskyRun
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget, QVBoxLayout
import logging
from typing import List

from xicam.gui.bluesky.central import CentralWidget

logger = logging.getLogger("BlueskyPlugin")


class SearchingCatalogController(QWidget):
    """
    Displays code that lets a user select from a root catalog.


    """

    sigOpen = Signal(object)
    sigSelectedRun = Signal([list])
    sigPreview = Signal(BlueskyRun)
    catalog = None

    def __init__(self, root_catalog):
        """
          The root catalog can be a single catalog or a collection of catalogs.
            Check out https://nsls-ii.github.io/databroker/v2/user/index.htmll#find-a-catalog. This
            class is intended to guide a user to select a catalog from the list in root catalog,
            then inspect and possibly open a selected run catalog.

        :param root_catalog: top level catalog that might contain more catalogs
        """
        super(SearchingCatalogController, self).__init__()

        layout = QVBoxLayout()
        self.setLayout(layout)

        layout.setContentsMargins(0, 0, 0, 0)
        self.centralWidget = CentralWidget(menuBar=None, catalog=root_catalog)

        def emit_opened_catalogs(name, catalogs: List[BlueskyRun]):
            """Emit each selected and opened catalog to sigOpen"""
            [self.sigOpen.emit(item) for item in catalogs]

        def preview_entry(name, catalog: BlueskyRun):
            self.sigPreview.emit(catalog)

        # connect the open_entries in the search model to sigOpen
        self.centralWidget.search_model.open_entries.connect(emit_opened_catalogs)
        self.centralWidget.search_model.preview_entry.connect(preview_entry)
        self.centralWidget.summary_widget.open.connect(emit_opened_catalogs)
        layout.addWidget(self.centralWidget)


class DatabrokerCatalogPlugin(CatalogPlugin):
    """
    Plugin for providing a user interface for browsing and selecting databroker
    catalogs. Users can select one or more catalogs in a list, which will signal
    to xi-cam sigOpen.
    """

    # set name here so that it appears in the catalog dropdown
    name = "Databroker"

    def __new__(cls):
        # importing catalog gives us #a catalog of pre-configured catalogs YAML files
        # locations by default might look something like:
        #  'USER_HOME/.local/share/intake', 'PYTHON_ENV/share/intake'
        from databroker import catalog

        catalog.controller = SearchingCatalogController(catalog)
        catalog.view = QListView()
        catalog.name = "Databroker"
        return catalog



import pytest
from pytestqt import qtbot

def test_IFittableModelPlugin(qtbot):
    from ..fittablemodelplugin import Fittable1DModelPlugin
    import numpy as np
    from astropy.modeling.fitting import LevMarLSQFitter
    from astropy.modeling import Parameter

    # Below example copied from AstroPy for demonstration; Gaussian1D is already a member of astropy's models
    class Gaussian1D(Fittable1DModelPlugin):
        amplitude = Parameter("amplitude")
        mean = Parameter("mean")
        stddev = Parameter("stddev")

        @staticmethod
        def evaluate(x, amplitude, mean, stddev):
            """
            Gaussian1D model function.
            """
            return amplitude * np.exp(-0.5 * (x - mean) ** 2 / stddev ** 2)

        @staticmethod
        def fit_deriv(x, amplitude, mean, stddev):
            """
            Gaussian1D model function derivatives.
            """

            d_amplitude = np.exp(-0.5 / stddev ** 2 * (x - mean) ** 2)
            d_mean = amplitude * d_amplitude * (x - mean) / stddev ** 2
            d_stddev = amplitude * d_amplitude * (x - mean) ** 2 / stddev ** 3
            return [d_amplitude, d_mean, d_stddev]

    # Generate fake data
    np.random.seed(0)
    x = np.linspace(-5.0, 5.0, 200)
    m_ref = Gaussian1D(amplitude=2.0, mean=1, stddev=3)
    from astropy.modeling.models import Gaussian1D

    Gaussian1D()(x)
    y = m_ref(x) + np.random.normal(0.0, 0.1, x.shape)

    # Fit model to data
    m_init = Gaussian1D()

    fit = LevMarLSQFitter()
    m = fit(m_init, x, y)

    assert round(m.amplitude.value) == 2
    assert round(m.mean.value) == 1
    assert round(m.stddev.value) == 3



from . import GUIPlugin, GUILayout
from pyqtgraph.parametertree import Parameter, ParameterTree
from qtpy.QtWidgets import *
from qtpy.QtGui import *


class _EZPlugin(GUIPlugin):
    def __init__(self, name, toolbuttons=None, parameters=None, appendcatalog=None, centerwidget=None, bottomwidget=None):
        self.name = name

        self.centerwidget = centerwidget() if callable(centerwidget) else centerwidget
        self.bottomwidget = bottomwidget() if callable(bottomwidget) else bottomwidget
        self.rightwidget = ParameterTree()
        self.toolbar = QToolBar()

        if parameters:
            self.parameters = Parameter(name="Params", type="group", children=parameters)
            self.rightwidget.setParameters(self.parameters, showTop=False)

        if toolbuttons:
            for toolbutton in toolbuttons:
                self.addToolButton(*toolbutton)

        if appendcatalog:
            self.appendCatalog = appendcatalog

        self.stages = {
            "EZPlugin": GUILayout(self.centerwidget, right=self.rightwidget, bottom=self.bottomwidget, top=self.toolbar)
        }

        super(_EZPlugin, self).__init__()
        _EZPlugin.instance = self

    def setImage(self, data):
        self.centerwidget.setImage(data)

    def plot(self, *args, **kwargs):
        self.bottomwidget.plot(*args, **kwargs)

    def addParameter(self, **kwargs):
        self.rightwidget.addParameters(Parameter(**kwargs))

    def addToolButton(self, icon, method, text=None):
        tb = QAction(QIcon(icon), text, self.toolbar)
        tb.triggered.connect(method)
        self.toolbar.addAction(tb)


def EZPlugin(name="TestPlugin", toolbuttons=None, parameters=None, appendcatalog=None, centerwidget=None, bottomwidget=None):
    """
    Quickly create a custom Xi-cam plugin.

    This function provides an easy-to-use interface for creating a customized
    Xi-cam plugin.

    Parameters
    ----------
    name : str
        The name of the plugin to create (the default is 'TestPlugin').
    toolbuttons : List[Tuple[str, Callable, (str)]]
        A list of tool buttons to create. A tool button will implement some
        action when the tool button is clicked. See the `Notes` section
        for more information.
    parameters : List[pyqtgraph.parametertree.Parameter]
        TODO
    appendheadertest
        TODO
    centerwidget : QWidget
        The widget displayed in the center of the plugin's GUI (the default is
        None, which creates an pyqtgraph.ImageView).
    bottomwidget : QWidget
        The widget displayed at the bottom (below the center) of the plugin's
        GUI (the default is None, which creates a pyqtgraph.PlotWidget).

    Returns
    -------
    EZPlugin derived type
        Returns an EZPlugin derived object of type `name`.

    Notes
    -----
    The `toolbutton` parameter is an iterable of tool buttons. Each
    tool button is an iterable with the following contents: the location of
    the icon for the button, an action (method) that is invoked when the button
    is triggered, and an optional text string for the button (e.g. when the
    mouse hovers over the button).

    """
    import pyqtgraph as pg

    if centerwidget is None:
        centerwidget = pg.ImageView
    if bottomwidget is None:
        bottomwidget = pg.PlotWidget
    return type(
        name,
        (_EZPlugin,),
        {
            "__init__": lambda self: _EZPlugin.__init__(
                self, name, toolbuttons, parameters, appendcatalog, centerwidget, bottomwidget
            )
        },
    )



from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from xicam.core import msg
from xicam.core.threads import invoke_in_main_thread, QThreadFuture


class SummaryWidget(QWidget):
    open = Signal([str, list])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.uid_label = QLabel()
        self.open_individually_button = QPushButton("Open")
        self.open_individually_button.setEnabled(False)
        self.open_individually_button.clicked.connect(self._open_individually)
        self.copy_uid_button = QPushButton("Copy UID to Clipboard")
        self.copy_uid_button.setEnabled(False)
        self.copy_uid_button.clicked.connect(self._copy_uid)
        self.streams = QLabel()
        self.entries = []

        uid_layout = QHBoxLayout()
        uid_layout.addWidget(self.uid_label)
        uid_layout.addWidget(self.copy_uid_button)
        layout = QVBoxLayout()
        layout.addWidget(self.open_individually_button)
        layout.addLayout(uid_layout)
        layout.addWidget(self.streams)
        self.setLayout(layout)

        self._tab_titles = ()

    def cache_tab_titles(self, titles):
        self._tab_titles = titles

    def _copy_uid(self):
        QApplication.clipboard().setText(self.uid)

    def _open_individually(self):
        for entry in self.entries:
            self.open.emit(None, [entry])

    def call_entry(self, entries):
        (entry,) = entries
        return entry()

    def set_entries(self, entries):
        try:
            self.entries.clear()
            self.entries.extend(entries)
            if not entries:
                self.uid_label.setText("")
                self.streams.setText("")
                self.copy_uid_button.setEnabled(False)
                self.open_individually_button.setEnabled(False)
            elif len(entries) == 1:
                future = QThreadFuture(self.call_entry, entries, callback_slot=self.call_entries_finished, showBusy=True)
                future.start()
            else:
                self.uid_label.setText("(Multiple Selected)")
                self.streams.setText("")
                self.copy_uid_button.setEnabled(False)
                self.open_individually_button.setText("Open individually")
                self.open_individually_button.setEnabled(True)
        except Exception as e:
            msg.logError(e)
            msg.showMessage("Unable to get data about selected run")

    def call_entries_finished(self, run):
        invoke_in_main_thread(self._call_entries_result, run)

    def _call_entries_result(self, run):
        self.uid = run.metadata["start"]["uid"]
        self.uid_label.setText(self.uid[:8])
        self.copy_uid_button.setEnabled(True)
        self.open_individually_button.setEnabled(True)
        self.open_individually_button.setText("Open")
        num_events = (run.metadata["stop"] or {}).get("num_events")
        if num_events:
            self.streams.setText("Streams:\n" + ("\n".join(f"{k} ({v} Events)" for k, v in num_events.items())))
        else:
            # Either the RunStop document has not been emitted yet or was never
            # emitted due to critical failure or this is an old document stream
            # from before 'num_events' was added to the schema. Get the list of
            # stream names another way, and omit the Event count.
            self.streams.setText("Streams:\n" + ("\n".join(list(run))))



import inspect

from PyQt5.QtGui import QCursor, QDrag, QPixmap, QRegion
from PyQt5.QtWidgets import QWidget, QTabWidget
from PyQt5.QtCore import Qt, QMimeData, QObject, QPoint
from traitlets import HasTraits, TraitType
from traitlets.config import Configurable

# These classes integrate Qt and traitlets so that we can subclass both.
# They are copied from Jupyter's qtconsole.util package. The only edits made
# here are removing PY2 support.


MetaHasTraits = type(HasTraits)
MetaQObject = type(QObject)


class MetaQObjectHasTraits(MetaQObject, MetaHasTraits):
    """ A metaclass that inherits from the metaclasses of HasTraits and QObject.
    Using this metaclass allows a class to inherit from both HasTraits and
    QObject. Using SuperQObject instead of QObject is highly recommended. See
    QtKernelManager for an example.
    """

    def __new__(mcls, name, bases, classdict):
        # FIXME: this duplicates the code from MetaHasTraits.
        # I don't think a super() call will help me here.
        for k, v in classdict.items():
            if isinstance(v, TraitType):
                v.name = k
            elif inspect.isclass(v):
                if issubclass(v, TraitType):
                    vinst = v()
                    vinst.name = k
                    classdict[k] = vinst
        cls = MetaQObject.__new__(mcls, name, bases, classdict)
        return cls

    def __init__(mcls, name, bases, classdict):
        # Note: super() did not work, so we explicitly call these.
        MetaQObject.__init__(mcls, name, bases, classdict)
        MetaHasTraits.__init__(mcls, name, bases, classdict)


def superQ(QClass):
    """ Permits the use of super() in class hierarchies that contain Qt classes.
    Unlike QObject, SuperQObject does not accept a QObject parent. If it did,
    super could not be emulated properly (all other classes in the heierarchy
    would have to accept the parent argument--they don't, of course, because
    they don't inherit QObject.)
    This class is primarily useful for attaching signals to existing non-Qt
    classes. See QtKernelManagerMixin for an example.
    """

    class SuperQClass(QClass):
        def __new__(cls, *args, **kw):
            # We initialize QClass as early as possible. Without this, Qt complains
            # if SuperQClass is not the first class in the super class list.
            inst = QClass.__new__(cls)
            QClass.__init__(inst)
            return inst

        def __init__(self, *args, **kw):
            # Emulate super by calling the next method in the MRO, if there is one.
            mro = self.__class__.mro()
            for qt_class in QClass.mro():
                mro.remove(qt_class)
            next_index = mro.index(SuperQClass) + 1
            if next_index < len(mro):
                mro[next_index].__init__(self, *args, **kw)

    return SuperQClass


SuperQObject = superQ(QObject)
ConfigurableQObject = MetaQObjectHasTraits("NewBase", (Configurable, SuperQObject), {})



from pytestqt import qtbot

def test_logIntensity(qtbot):
    from xicam.gui.widgets.imageviewmixins import LogScaleIntensity
    import numpy as np



    windows = []

    data1 = np.fromfunction(lambda x, y: np.exp((x ** 2 + y ** 2) / 10000.0), (100, 100)) - 2
    data2 = np.random.random((100, 100))
    data3 = np.random.random((100, 100)) * 1000 - 2
    data3[:10, :10] = np.random.random((10, 10)) * 10 - 2

    for data in [data1, data2, data3]:
        w = LogScaleIntensity()
        w.setImage(data)
        w.show()
        windows.append(w)



if __name__ == "__main__":
    test_logIntensity()



import pytest

from xicam.core import execution
from xicam.core.execution import localexecutor
from xicam.core.execution.workflow import Graph, Workflow
from xicam.plugins import OperationPlugin
from xicam.plugins.operationplugin import output_names, operation, display_name
from pyqtgraph.parametertree import Parameter


@pytest.fixture()
def graph():
    return Graph()


@pytest.fixture()
def square_op():
    @operation
    def square(n: int) -> int:
        return n * n

    return square()


@pytest.fixture()
def sum_op():
    @operation
    def sum(n1: int, n2: int) -> int:
        return n1 + n2

    return sum()


@pytest.fixture()
def negative_op():
    @operation
    def negative(num: int) -> int:
        return -1 * num

    return negative()


@pytest.fixture()
def simple_workflow(square_op, sum_op):
    from xicam.core.execution.workflow import Workflow

    wf = Workflow()

    square = square_op
    square2 = square_op.__class__()
    square2.filled_values["n1"] = 2

    wf.add_operation(square)
    wf.add_operation(square2)
    wf.add_operation(sum_op)
    wf.add_link(square, sum_op, "square", "n1")
    wf.add_link(square2, sum_op, "square", "n2")

    return wf


@pytest.fixture()
def custom_parameter_op():
    class CustomParameterOp(OperationPlugin):
        def __init__(self):
            super(CustomParameterOp, self).__init__()
            self.value = False

        def _func(self):
            return self.value

        def as_parameter(self):
            return [{'name':'test', 'type':'bool'}]

        def wireup_parameter(self, parameter:Parameter):

            parameter.child('test').sigValueChanged.connect(lambda value: print(value))

    return CustomParameterOp


@pytest.fixture()
def custom_parameter_workflow(custom_parameter_op):
    from xicam.core.execution.workflow import Workflow

    wf = Workflow()

    custom_parameter_op = custom_parameter_op()

    wf.add_operation(custom_parameter_op)
    return wf


import argparse
from functools import lru_cache


class _RaisingAruementParser(argparse.ArgumentParser):
    """Raise instead of exit on failure to parse arguments

    This is because we need to parse at import in xicam.plugin but
    should be forgiving of importing in contexts other than the
    canonical xicam application.

    We must parse on import because we use CLI flags to controls if
    camart is enabled in the `__init__` of
    xicam.plugin.XicamPluginManager which we then instatiate on import.

    This singleton is then imported many other places.

    This class should be removed when we remove the need to parse on
    import.

    """

    _orig_error = argparse.ArgumentParser.error

    def error(self, message):
        raise RuntimeError(message)


@lru_cache()
def parse_args(exit_on_fail=True):
    parser = _RaisingAruementParser()
    parser.add_argument("-v", "--verbose", dest="verbose", action="count", help="increase output verbosity", default=0)
    parser.add_argument(
        "--no-cammart", dest="nocammart", action="store_true", help="disable cammart and sandboxed environment features"
    )
    parser.add_argument(
        "--blacklist",
        dest="blacklist",
        action="append",
        help="prevent Xi-cam from loading a plugin by name",
        type=str,
        default=[],
    )
    parser.add_argument("--no-splash", dest="nosplash", action="store_true", help="skip the Xi-cam splash screen")
    try:
        return parser.parse_args()
    except RuntimeError as re:
        if exit_on_fail:
            parser._orig_error(str(re))
        else:
            raise



from pytestqt import qtbot


def test_IDataSourcePlugin(qtbot):
    from xicam.plugins.dataresourceplugin import DataResourcePlugin, DataSourceListModel

    class SpotDataResourcePlugin(DataResourcePlugin):
        def __init__(
            self,
            user="anonymous",
            password="",
            query="skipnum=0&sortterm=fs.stage_date&sorttype=desc&search=end_station=bl832",
        ):
            scheme = "https"
            host = "portal-auth.nersc.gov"
            path = "als/hdf/search"
            config = {"scheme": scheme, "host": host, "path": path, "query": query}
            super(SpotDataResourcePlugin, self).__init__(flags={"canPush": False}, **config)
            from requests import Session

            self.session = Session()
            self.refresh()
            try:
                self.session.post("https://newt.nersc.gov/newt/auth", {"username": user, "password": password})
            except ConnectionError:
                pass  # TODO: Something

        def columnCount(self, index=None):
            return len(self._data[0])

        def rowCount(self, index=None):
            return len(self._data)

        def data(self, index, role):
            from qtpy.QtCore import Qt, QVariant

            if index.isValid() and role == Qt.DisplayRole:
                return QVariant(self._data[index.row()]["name"])
            else:
                return QVariant()

                # TODO: remove qtcore dependence

        def refresh(self):
            self._data = []
            try:
                r = self.session.get(
                    "https://portal-auth.nersc.gov/als/hdf/search?skipnum=0&limitnum=10&sortterm=fs.stage_date&sorttype=desc&search=end_station=bl832"
                )
                self._data = eval(r.content.replace(b"false", b"False"))
            except ConnectionError:
                pass  # TODO: SOMethING?

    # app = makeapp()
    # from qtpy.QtWidgets import QListView
    #
    # # TODO: handle password for testing
    # spot = DataSourceListModel(SpotDataResourcePlugin())
    #
    # lv = QListView()
    # lv.setModel(spot)
    # lv.show()
    # mainloop()

    spot = SpotDataResourcePlugin()
    assert spot.rowCount()
    qtbot.addWidget(spot)



import itertools
from .plugin import PluginType
from qtpy.QtWidgets import QListView, QWidget, QVBoxLayout
from qtpy.QtCore import Signal, QAbstractItemModel, QModelIndex, Qt, QObject
from intake.catalog.base import Catalog
from xicam.core import msg
from collections import OrderedDict


# from intake_bluesky.in_memory import SafeLocalCatalogEntry


class CatalogModel(QAbstractItemModel):
    """
    This model binds to a Catalog, and represents its contents in a paginated way, configurable by `pagination_size`

    Its expected use is either
    - uniquely bound to a static Catalog
    - dynamically rebound to catalogs (i.e. as a datasource is filtered) with `.setCatalog`
    """

    pagination_size = 10

    def __init__(self, catalog: Catalog):
        self.catalog = catalog
        super(CatalogModel, self).__init__()

        # Note: both of these caches are used so that indexing by row is performant
        # A cache of the RunCatalogs seen by this model
        self._cache = []

        # For iterating over the catalog as needed
        self._run_iterator = self.catalog.__iter__()

        # This model will add items to itself at the request any view
        self._rowcount = 0

    def index(self, row, column, parent):
        return self.createIndex(row, column)

    def parent(self, index):
        return QModelIndex()

    def rowCount(self, index):
        return self._rowcount

    def columnCount(self, index):
        return 1

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            return self._cache[index.row()].name

    def canFetchMore(self, parent):
        if parent.isValid():
            return False
        return self._rowcount < len(self.catalog)

    def fetchMore(self, parent):
        if parent.isValid():
            return

        # prevent fetching more items than the catalog currently has
        to_fetch = min(len(self.catalog) - self._rowcount, self.pagination_size)

        # print(f'Fetching next {to_fetch} runs')

        # pre-fetch more uids from the datasource
        new_uids = itertools.islice(self._run_iterator, to_fetch)

        # fetch more RunCatalogs from the datasource
        self._cache.extend(self.catalog[uid] for uid in new_uids)

        self.beginInsertRows(QModelIndex(), self._rowcount, self._rowcount + to_fetch)
        self._rowcount += to_fetch  # Tell the model it now has more rows
        self.endInsertRows()

    # NOTE: the following methods are expected to be called by an external controller

    def setCatalog(self, catalog):
        self.catalog = catalog
        self.reset()

    def reset(self):
        self._cache = []
        self._rowcount = 0
        self._run_iterator = self.catalog.__iter__()


class CatalogController(QWidget):
    # TODO: Make a desicion what we want these signal objects to be
    sigOpen = Signal(object)
    # TODO: Make sure these are emitted / connected
    sigPreview = Signal(object)

    sigOpenPath = Signal(str)
    sigOpenExternally = Signal(str)
    # TODO: Emit original / new str
    sigLocationChanged = Signal()

    def __init__(self, view, parent=None):
        super(CatalogController, self).__init__(parent=parent)
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(view)

        self.view = view

        # Setup signal emissions
        view.doubleClicked.connect(self.open)

    def open(self, _):
        indexes = self.view.selectionModel().selectedRows()

        runs = [self.view.model().catalog[self.view.model().data(index)] for index in indexes]

        for run in runs:
            self.sigOpen.emit(run)


class CatalogPlugin(Catalog, PluginType):
    is_singleton = False
    needs_qt = True
    name = ""

    model = CatalogModel
    view = QListView
    controller = CatalogController

    def __init__(self, *args, **kwargs):
        super(CatalogPlugin, self).__init__()
        self.setup()

    def setup(self):
        import inspect

        if inspect.isclass(self.model):
            self.model = self.model(self)

        if inspect.isclass(self.view):
            self.view = self.view()
            self.view.setModel(self.model)

        if inspect.isclass(self.controller):
            self.controller = self.controller(self.view)


if __name__ == "__main__":
    from intake_bluesky.jsonl import BlueskyJSONLCatalog
    import glob

    class JSONLCatalogPlugin(BlueskyJSONLCatalog, CatalogPlugin):
        ...

    from qtpy.QtWidgets import QApplication

    qapp = QApplication([])

    paths = glob.glob("/home/rp/data/Catalog Sample/abc/*.jsonl")
    model = JSONLCatalogPlugin(paths).model

    view = QListView()
    view.setModel(model)

    view.show()
    qapp.exec_()



from pytestqt import qtbot
import pytest
from xicam.plugins.operationplugin import (
    display_name,
    fixed,
    input_names,
    limits,
    opts,
    output_names,
    output_shape,
    plot_hint,
    units,
    visible,
    ValidationError,
    operation,
)
from xicam.core.execution import Workflow
from xicam.gui.widgets.linearworkfloweditor import WorkflowEditor
from xicam.core.tests.workflow_fixtures import square_op, sum_op, custom_parameter_op, custom_parameter_workflow, simple_workflow


def test_simple(simple_workflow: Workflow, qtbot):
    workflow_editor = WorkflowEditor(simple_workflow)
    workflow_editor.show()
    qtbot.addWidget(workflow_editor)


def test_custom_parameter(custom_parameter_workflow: Workflow, qtbot):
    workflow_editor = WorkflowEditor(custom_parameter_workflow)
    workflow_editor.show()
    qtbot.addWidget(workflow_editor)
    workflow_editor.workflowview.selectRow(0)  # Note: models is empty here because pluginmanger hasn't finished load yet



from pyqtgraph.parametertree import ParameterTree
from pyqtgraph.parametertree.parameterTypes import Parameter, GroupParameter
from collections import deque, OrderedDict, defaultdict
from qtpy.QtGui import QStandardItem, QStandardItemModel
from qtpy.QtCore import QItemSelectionModel, Signal
import sys
import uuid
import datetime
import numpy as np
import warnings
from typing import Iterable, Sequence
from xicam.core import msg
from xicam.core.data import NonDBHeader
from xicam.gui.patches.PyQtGraph import CounterGroupParameter, LazyGroupParameter

# TODO: map list to groupparameter

# TODO: suggest integration of type mapping into pyqtgraph
typemap = {
    int: "int",
    float: "float",
    np.float64: "float",
    str: "str",
    dict: "lazygroup",
    OrderedDict: "lazygroup",
    list: "lazygroup",
    type(None): None,
    tuple: "lazygroup",
    datetime.datetime: "str",
    np.float: "float",
    np.ndarray: "ndarray",
    np.int: "int",
}


class MetadataWidgetBase(ParameterTree):
    def __init__(self, *args, **kwargs):
        self.excludedkeys = kwargs.get("excludedkeys")
        if "excludedkeys" in kwargs:
            del kwargs["excludedkeys"]

        super(MetadataWidgetBase, self).__init__(*args, **kwargs)
        LazyGroupParameter.itemClass.initialize_treewidget(self)
        self.kwargs = kwargs

    def insert(self, doctype: str, document, groups: dict):
        if doctype == "start":
            for group in groups.values():
                group.clearChildren()

        try:
            new_children = MetadataView._from_dict(document, self.excludedkeys)
        except Exception as ex:
            msg.logError(ex)
            print(f"failed to make children for {doctype}")
        else:
            # TODO: add responsive design to uid display
            group = groups[doctype]
            group.addChildren(
                [
                    GroupParameter(
                        name=document["uid"][:6], value=None, type=None, children=new_children, expanded=False, readonly=True
                    )
                ]
            )

    @staticmethod
    def _from_dict(metadata: Iterable, excludedkeys=None):
        if isinstance(metadata, Sequence):
            metadata = OrderedDict(enumerate(metadata))

        metadata = MetadataView._strip_reserved(metadata, excludedkeys)
        children = []
        for key, value in metadata.items():
            subchildren = []
            if typemap.get(type(value), None) in ["group", "lazygroup"]:
                subchildren = MetadataView._from_dict(value)
                key = f"{str(key)} {type(value)}"
                value = None
            try:
                children.append(
                    Parameter.create(
                        name=str(key),
                        value=value,
                        type=typemap[type(value)],
                        children=subchildren,
                        expanded=False,
                        readonly=True,
                    )
                )
            except KeyError:
                warnings.warn(f"Failed to display a {type(value)}: {value}")
                children.append(
                    Parameter.create(
                        name=str(key),
                        value=repr(value),
                        type=typemap[str],
                        children=subchildren,
                        expanded=False,
                        readonly=True,
                    )
                )
        return children

    @staticmethod
    def _strip_reserved(metadata: dict, excludedkeys=None):
        if not excludedkeys:
            excludedkeys = []
        metadata = metadata.copy()
        for key in excludedkeys:
            if key in metadata:
                del metadata[key]
        return metadata


class MetadataWidget(MetadataWidgetBase):
    def __init__(self, *args, **kwargs):
        super(MetadataWidget, self).__init__(*args, **kwargs)
        self.header = None  # type: HeaderParameter
        self.groups = dict()
        self.reset()

    def doc_consumer(self, name, doc):
        if name == "start":
            self.header.setName(doc["uid"])
        super(MetadataWidget, self).insert(name, doc, self.groups)

    def reset(self):
        self.header = HeaderParameter(name=" ")
        self.groups = {group.name(): group for group in self.header.children()}
        self.setParameters(self.header)


class MetadataView(MetadataWidgetBase):
    sigUpdate = Signal()

    def __init__(self, headermodel: QStandardItemModel, selectionmodel: QItemSelectionModel, *args, **kwargs):
        super(MetadataView, self).__init__(*args, **kwargs)
        self._seen = set()
        self._last_uid = None
        self._thread_id = "MetadataView" + str(uuid.uuid4())
        self.headermodel = headermodel
        self.selectionmodel = selectionmodel
        self.selectionmodel.currentChanged.connect(self.update)
        self.sigUpdate.connect(self.update)
        self.update()

    def update(self):
        index = self.selectionmodel.currentIndex()
        if not index.isValid():
            return

        header = self.headermodel.itemFromIndex(index).header  # type: NonDBHeader

        if header.startdoc["uid"] != self._last_uid:
            self._seen = set()

        if not isinstance(header, HeaderBuffer):
            headerbuffer = HeaderBuffer(
                header.startdoc["uid"], HeaderParameter(name=header.startdoc["uid"]), docs=header.documents()
            )

        param = headerbuffer.param
        groups = param.groups

        # TODO: make compatible with actively streaming header

        # filter out documents already emitted in the stream
        for doctype, document in header.stream():
            if document["uid"] in self._seen:
                continue

            self._seen.add(document["uid"])
            self.insert(doctype, document, groups)

        if headerbuffer.uid != self._last_uid:
            self._last_uid = headerbuffer.uid
            # try:
            self.setParameters(param, showTop=False)


class HeaderParameter(GroupParameter):
    def __init__(self, *args, **kwargs):
        super(HeaderParameter, self).__init__(*args, **kwargs)

        self.groups = {
            "start": CounterGroupParameter(name="start", title="Start", expanded=False),
            "descriptor": CounterGroupParameter(name="descriptor", title="Descriptors", expanded=False),
            "event": CounterGroupParameter(name="event", title="Events", expanded=False),
            "stop": CounterGroupParameter(name="stop", title="Stop", expanded=False),
        }
        self.addChildren(self.groups.values())


class HeaderBuffer:
    def __init__(self, uid, param: GroupParameter, docs=None):
        self.uid = uid
        self.buf = deque()
        self.param = param

        for name, doc in docs:
            self.buf.append((name, doc))

    def stream(self, *args):
        yield from self.buf


class MDVConusumer(MetadataView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._buffers = {}
        self._descriptor_map = {}

    def doc_consumer(self, name, doc):
        if name == "start":
            uid = doc["uid"]
            item = QStandardItem(doc["uid"])
            item.header = self._buffers[uid] = HeaderBuffer(uid, HeaderParameter(name="uid"))
            self.headermodel.appendRow(item)
            self.selectionmodel.setCurrentIndex(self.headermodel.indexFromItem(item), QItemSelectionModel.ClearAndSelect)
        elif name == "descriptor":
            uid = self._descriptor_map[doc["uid"]] = doc["run_start"]
        elif name == "event":
            uid = self._descriptor_map[doc["descriptor"]]
        elif name == "stop":
            uid = doc["run_start"]
        else:
            raise ValueError

        buffer = self._buffers[uid].buf  # type: deque
        buffer.append((name, doc))
        self.sigUpdate.emit()

    def show_row_n(self, n):
        item = self.headermodel.item(n)
        if item is None:
            return
        self.selectionmodel.setCurrentIndex(self.headermodel.indexFromItem(item), QItemSelectionModel.ClearAndSelect)
        self.update()


if __name__ == "__main__":
    import os

    os.environ["OPHYD_CONTROL_LAYER"] = "caproto"
    from qtpy import QtWidgets
    from qtpy import QtCore
    from qtpy import QtGui
    from mily.runengine import spawn_RE
    from mily.widgets import ControlGui, Count, Scan1D, MotorSelector, DetectorSelector, MISpin
    import bluesky.plans as bp
    from ophyd.sim import hw
    import matplotlib

    matplotlib.interactive(True)

    class MDVWithButtons(QtWidgets.QWidget):
        def __init__(self, mdv, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.mdv = mdv

            w_layout = QtWidgets.QVBoxLayout()
            w_layout.addWidget(mdv)

            button_layout = QtWidgets.QHBoxLayout()
            w_layout.addLayout(button_layout)

            self.spinner = spinner = MISpin("run")
            w_layout.addWidget(spinner)
            spinner.setRange(0, 100)

            def on_change(n):
                self.mdv.show_row_n(n)

            spinner.valueChanged.connect(on_change)
            self.setLayout(w_layout)

        def doc_consumer(self, name, doc):
            self.mdv.doc_consumer(name, doc)
            self.spinner.setRange(0, self.mdv.headermodel.rowCount() - 1)

    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication([b"Live data demo"])
        app.lastWindowClosed.connect(app.quit)

    headermodel = QtGui.QStandardItemModel()
    selectionmodel = QtCore.QItemSelectionModel(headermodel)

    metadataview = MDVConusumer(headermodel, selectionmodel)
    view_box = MDVWithButtons(metadataview)

    hw = hw()
    hw.motor.set(15)
    hw.motor.delay = 0.1
    hw.motor1.delay = 0.2
    hw.motor2.delay = 0.3

    hw.det.kind = "hinted"
    hw.det1.kind = "hinted"
    hw.det2.kind = "hinted"

    RE, queue, thread, teleport = spawn_RE()

    cg = ControlGui(
        queue,
        teleport,
        Count("Count", bp.count, DetectorSelector(detectors=[hw.det, hw.det1, hw.det2])),
        Scan1D(
            "1D absolute scan",
            bp.scan,
            MotorSelector([hw.motor, hw.motor1, hw.motor2]),
            DetectorSelector(detectors=[hw.det, hw.det1, hw.det2]),
        ),
        Scan1D(
            "1D relative scan",
            bp.rel_scan,
            MotorSelector([hw.motor, hw.motor1, hw.motor2]),
            DetectorSelector(detectors=[hw.det, hw.det1, hw.det2]),
        ),
        live_widget=view_box,
    )

    cg.show()
    sys.exit(app.exec_())



from dask.diagnostics import Profiler, ResourceProfiler, CacheProfiler
from dask.diagnostics import visualize
from xicam.core import msg
from .workflow import Workflow
from appdirs import user_config_dir
import distributed


class DaskExecutor(object):
    def __init__(self):
        super(DaskExecutor, self).__init__()
        self.client = None

    def execute(self, wf: Workflow, client=None):
        if not wf.operations:
            return {}

        if client is None:
            if self.client is None:
                self.client = distributed.Client()
            client = self.client

        dask_graph, end_task_ids = wf.as_dask_graph()

        # with Profiler() as prof, ResourceProfiler(dt=0.25) as rprof, CacheProfiler() as cprof:
        result = client.get(dask_graph, end_task_ids)

        msg.logMessage("result:", result, level=msg.DEBUG)
        # path = user_config_dir('xicam/profile.html')
        # visualize([prof, rprof, cprof], show=False, file_path=path)
        # msg.logMessage(f'Profile saved: {path}')

        wf.lastresult = result

        return result



import pytest
from pytestqt import qtbot


def test_EZPlugin(qtbot):
    from xicam.plugins import EZPlugin
    from xicam.gui.static import path
    from xicam.core.data import NonDBHeader

    def runtest():
        import numpy as np

        img = np.random.random((100, 100, 100))
        EZTest.instance.setImage(img)

        hist = np.histogram(img, 100)
        EZTest.instance.plot(hist[1][:-1], hist[0])

    def appendcatalog(header: NonDBHeader):
        img = header.meta_array(list(header.fields())[0])
        EZTest.instance.setImage(img)

    EZTest = EZPlugin(
        name="EZTest",
        toolbuttons=[(str(path("icons/calibrate.png")), runtest)],
        parameters=[{"name": "Test", "value": 10, "type": "int"}, {"name": "Fooo", "value": True, "type": "bool"}],
        appendcatalog=appendcatalog,
    )
    return EZTest




from pyqtgraph import ROI, PolyLineROI, Point
from pyqtgraph.graphicsItems.ROI import Handle, RectROI, LineROI
from qtpy.QtCore import QRectF, QPointF, Qt
from qtpy.QtGui import QColor, QPainter, QPainterPath, QBrush
import numpy as np
from itertools import count
from xicam.plugins import OperationPlugin


from pyqtgraph.parametertree import Parameter, parameterTypes
import pyqtgraph as pg


class ROIProcessingPlugin(OperationPlugin):
    name = 'ROI'
    output_names = ('ROI', 'data')

    def __init__(self, ROI: ROI):
        super(ROIProcessingPlugin, self).__init__()
        self.ROI = ROI
        self._param = None  # type: Parameter

        self.name = f"ROI #{self.ROI.index}"

    def _func(self, data, image):
        data = self.ROI.getLabelArray(data, image)
        roi = data.astype(np.bool)
        return roi, data

    @property
    def parameter(self):
        if not self._param:
            self._param = self.ROI.parameter()
        return self._param


class WorkflowableROI(ROI):
    def __init__(self, *args, **kwargs):
        super(WorkflowableROI, self).__init__(*args, **kwargs)
        self.process = ROIProcessingPlugin(self)
        self._param = None

    def parameter(self) -> Parameter:
        raise NotImplementedError


# MIXIN!~
# Now with 100% more ROI!
class BetterROI(WorkflowableROI):
    roi_count = count(0)
    index = None

    def __init__(self, *args, removable=True, **kwargs):
        # BetterROI removable by default
        super(BetterROI, self).__init__(*args, removable=removable, **kwargs)
        self.index = next(self.roi_count)
        self._restyle()
        # Remove the roi from the view when requested to be removed
        self.sigRemoveRequested.connect(lambda roi: self._viewBox().removeItem(roi))

    def _restyle(self):
        self.currentPen.setWidth(2)

        for handledict in self.handles:  # type: dict
            handle = handledict["item"]  # type: Handle
            handle.radius = handle.radius * 2
            handle.pen.setWidth(2)
            handle.buildPath()

    def hoverEvent(self, ev):
        hover = False
        if not ev.isExit():
            if ev.acceptDrags(Qt.LeftButton):
                hover = True
            for btn in [Qt.LeftButton, Qt.RightButton, Qt.MidButton]:
                if int(self.acceptedMouseButtons() & btn) > 0 and ev.acceptClicks(btn):
                    hover = True

        if hover:
            self.currentPen = pg.mkPen(255, 255, 0, width=2)
        else:
            self.currentPen = self.pen
        self.update()

    def valueChanged(self, sender, changes):
        for change in changes:
            setattr(self, change[0].name(), change[2])
        self.stateChanged()


class BetterPolyLineROI(BetterROI, PolyLineROI):
    def __repr__(self):
        return f"ROI #{self.index}"


class QCircRectF(QRectF):
    def __init__(self, center=(0.0, 0.0), radius=1.0, rect=None):
        self._scale = 1.0
        if rect is not None:
            self.center = rect.center()
            super(QCircRectF, self).__init__(rect)
        else:
            self.center = QPointF(*center)

            left = self.center.x() - radius
            top = self.center.y() - radius
            bottom = self.center.y() + radius
            right = self.center.x() + radius

            super(QCircRectF, self).__init__(QPointF(left, top), QPointF(right, bottom))

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = value
        self.radius *= value
        self.setLeft(self.center.x() - self._radius)
        self.setTop(self.center.y() - self._radius)
        self.setBottom(self.center.y() + self._radius)
        self.setRight(self.center.x() + self._radius)

    @property
    def radius(self):
        return (self.right() - self.left()) * 0.5

    @radius.setter
    def radius(self, radius):

        self.setLeft(self.center.x() - radius)
        self.setTop(self.center.y() - radius)
        self.setBottom(self.center.y() + radius)
        self.setRight(self.center.x() + radius)


class ArcROI(BetterROI):
    """
    A washer-wedge-shaped ROI for selecting q-ranges

    """

    def __init__(self, center, radius, **kwargs):
        # QtGui.QGraphicsRectItem.__init__(self, 0, 0, size[0], size[1])
        r = QCircRectF(center, radius)
        super(ArcROI, self).__init__(r.center, radius, **kwargs)
        # self.addRotateHandle([1.0, 0.5], [0.5, 0.5])
        # self.addScaleHandle([0.5*2.**-0.5 + 0.5, 0.5*2.**-0.5 + 0.5], [0.5, 0.5])

        self.startangle = 30
        self.arclength = 120

        self.aspectLocked = True
        self.translatable = False
        self.translateSnap = False
        self.removable = True

        self.center = center

        # only these values are in external space, others are internal (-.5,.5)
        self.innerradius = 0.5 * radius
        self.outerradius = radius
        self.thetawidth = 120.0
        self.thetacenter = 90.0

        self.innerhandle = self.addFreeHandle([0.0, self.innerradius / self.outerradius], [0, 0])
        self.outerhandle = self.addFreeHandle([0.0, 1], [0, 0])
        self.widthhandle = self.addFreeHandle(np.array([-0.433 * 2, 0.25 * 2]))

        self.path = None
        self._param = None  # type: Parameter
        self._restyle()

    def boundingRect(self):
        size = self.outerradius
        return QRectF(-size, -size, size * 2, size * 2).normalized()

    def movePoint(self, handle, pos, modifiers=Qt.KeyboardModifier(), finish=True, coords="parent"):
        super(ArcROI, self).movePoint(handle, pos, modifiers, finish, coords)

        # Set internal parameters
        if handle in [self.innerhandle, self.outerhandle]:
            self.innerradius = self.innerhandle.pos().length()
            self.outerradius = self.outerhandle.pos().length()
            self.thetacenter = self.outerhandle.pos().angle(Point(1, 0))

        elif handle is self.widthhandle:
            self.thetawidth = 2 * self.widthhandle.pos().angle(self.innerhandle.pos())

        self.handleChanged()

    def paint(self, p, opt, widget):

        # Enforce constraints on handles
        r2 = Point(np.cos(np.radians(self.thetacenter)), np.sin(np.radians(self.thetacenter)))  # chi center direction vector
        # constrain innerhandle to be parallel to outerhandle, and shorter than outerhandle
        self.innerhandle.setPos(r2 * self.innerradius)
        # constrain widthhandle to be counter-clockwise from innerhandle
        widthangle = np.radians(self.thetawidth / 2 + self.thetacenter)
        widthv = Point(np.cos(widthangle), np.sin(widthangle)) if self.thetawidth > 0 else r2
        # constrain widthhandle to half way between inner and outerhandles
        self.widthhandle.setPos(widthv * (self.innerradius + self.outerradius) / 2)
        # constrain handles to base values
        self.outerhandle.setPos(r2 * self.outerradius)

        pen = self.currentPen
        pen.setColor(QColor(0, 255, 255))

        p.setPen(pen)

        r = self.boundingRect()
        # p.drawRect(r)
        p.setRenderHint(QPainter.Antialiasing)

        p.scale(r.width(), r.height())  # workaround for GL bug

        centerangle = self.innerhandle.pos().angle(Point(1, 0))
        startangle = centerangle - self.thetawidth / 2
        endangle = centerangle + self.thetawidth / 2

        r = QCircRectF(radius=0.5)
        if self.innerradius < self.outerradius and self.thetawidth > 0:
            p.drawArc(r, -startangle * 16, -self.thetawidth * 16)

        radius = self.innerradius / self.outerradius / 2
        r = QCircRectF()
        r.radius = radius

        if self.innerradius < self.outerradius and self.thetawidth > 0:
            p.drawArc(r, -startangle * 16, -self.thetawidth * 16)

        pen.setStyle(Qt.DashLine)
        p.setPen(pen)

        p.drawLine(QPointF(0.0, 0.0), self.widthhandle.pos().norm() / 2)
        r1v = self.innerhandle.pos().norm()
        p.drawLine(QPointF(0.0, 0.0), (-1.0 * self.widthhandle.pos() + 2 * self.widthhandle.pos().dot(r1v) * r1v).norm() / 2)
        pen.setStyle(Qt.SolidLine)

        if self.innerradius < self.outerradius and self.thetawidth > 0:
            path = QPainterPath()
            path.moveTo((-1.0 * self.widthhandle.pos() + 2 * self.widthhandle.pos().dot(r1v) * r1v).norm() / 2)
            path.arcTo(r, -startangle, -self.thetawidth)  # inside
            path.lineTo(self.widthhandle.pos().norm() / 2)  # ? side
            path.arcTo(QCircRectF(radius=0.5), -endangle, self.thetawidth)  # outside
            path.lineTo((-1.0 * self.widthhandle.pos() + 2 * self.widthhandle.pos().dot(r1v) * r1v).norm() / 2)
            self.path = path
            p.fillPath(path, QBrush(QColor(0, 255, 255, 20)))

    def getArrayRegion(self, arr, img=None):
        """
        Return the result of ROI.getArrayRegion() masked by the arc shape
        of the ROI. Regions outside the arc are set to 0.
        """
        w = arr.shape[0]
        h = arr.shape[1]

        centerangle = self.outerhandle.pos().angle(Point(1, 0))
        startangle = centerangle - self.thetawidth / 2

        # generate an ellipsoidal mask
        mask = np.fromfunction(
            lambda x, y: (
                self.innerhandle.pos().length() < ((x - self.center[0]) ** 2.0 + (y - self.center[1]) ** 2.0) ** 0.5
            )
            & (((x - self.center[0]) ** 2.0 + (y - self.center[1]) ** 2.0) ** 0.5 < self.outerhandle.pos().length())
            & ((np.degrees(np.arctan2(y - self.center[1], x - self.center[0])) - startangle) % 360 > 0)
            & ((np.degrees(np.arctan2(y - self.center[1], x - self.center[0])) - startangle) % 360 < self.thetawidth),
            (w, h),
        )

        return arr * mask

    def shape(self):
        # (used for hitbox for menu)

        centerangle = self.innerhandle.pos().angle(Point(1, 0))
        startangle = centerangle - self.thetawidth / 2
        endangle = centerangle + self.thetawidth / 2
        r1v = self.innerhandle.pos().norm()

        # Draw out the path in external space
        path = QPainterPath()
        path.moveTo(-1.0 * self.widthhandle.pos() + 2 * self.widthhandle.pos().dot(r1v) * r1v)
        path.arcTo(QCircRectF(radius=self.innerradius), -startangle, -self.thetawidth)  # inside
        path.lineTo(self.widthhandle.pos())  # ? side
        path.arcTo(QCircRectF(radius=self.outerradius), -endangle, self.thetawidth)  # outside
        path.lineTo(-1.0 * self.widthhandle.pos() + 2 * self.widthhandle.pos().dot(r1v) * r1v)
        return path

    def parameter(self):
        if not self._param:
            self._param = parameterTypes.GroupParameter(
                name="Arc ROI",
                children=[
                    parameterTypes.SimpleParameter(
                        title="Q Minimum", name="innerradius", value=self.innerradius, type="float", units="Å⁻¹"
                    ),
                    parameterTypes.SimpleParameter(
                        title="Q Maximum", name="outerradius", value=self.outerradius, type="float", units="Å⁻¹"
                    ),
                    parameterTypes.SimpleParameter(
                        title="χ Width", name="thetawidth", value=self.thetawidth, type="float", units="°"
                    ),
                    parameterTypes.SimpleParameter(
                        title="χ Center", name="thetacenter", value=self.thetacenter, type="float", siSuffix="°"
                    ),
                ],
            )

            self._param.sigTreeStateChanged.connect(self.valueChanged)

        return self._param

    def handleChanged(self):
        self.parameter().child("innerradius").setValue(self.innerradius)
        self.parameter().child("outerradius").setValue(self.outerradius)
        self.parameter().child("thetawidth").setValue(self.thetawidth)
        self.parameter().child("thetacenter").setValue(self.thetacenter)


class RectROI(BetterROI, RectROI):
    def __init__(self, *args, pen=pg.mkPen(QColor(0, 255, 255)), **kwargs):
        super(RectROI, self).__init__(*args, pen=pen, **kwargs)
        self.handle = self.handles[0]

    def movePoint(self, handle, pos, modifiers=Qt.KeyboardModifier(), finish=True, coords="parent"):
        super(RectROI, self).movePoint(handle, pos, modifiers, finish, coords)

        self.width = self.handle["pos"].x() * self.size().x()
        self.height = self.handle["pos"].y() * self.size().y()

        self.handleChanged()

    def parameter(self) -> Parameter:
        if not self._param:
            self._param = parameterTypes.GroupParameter(
                name="Rectangular ROI",
                children=[
                    parameterTypes.SimpleParameter(title="Width", name="width", value=self.width, type="float", units="px"),
                    parameterTypes.SimpleParameter(
                        title="Height", name="height", value=self.height, type="float", units="px"
                    ),
                ],
            )

            self._param.sigTreeStateChanged.connect(self.valueChanged)

        return self._param

    def handleChanged(self):
        self.parameter().child("width").setValue(self.width)
        self.parameter().child("height").setValue(self.height)

    def getLabelArray(self, arr, img: pg.ImageItem = None):
        # TODO : make more generic for all rectangle ROIs, segmented (multi-labeled) and non-segmented (single-labeled)
        dim_0, dim_1 = arr.shape

        min_x = self.pos().x()
        min_y = self.pos().y()
        max_x = self.size().x() + min_x
        max_y = self.size().y() + min_y

        mask = np.zeros_like(arr)

        label_mask = np.fromfunction(
            lambda y, x: (x + 0.5 > min_x) & (x + 0.5 < max_x) & (y + 0.5 > min_y) & (y + 0.5 < max_y), (dim_0, dim_1)
        )
        mask[label_mask] = 1

        # Invert y
        # FIXME -- use image transform above with passed image item
        return mask[::-1, ::]


class LineROI(BetterROI, LineROI):
    def __init__(self, *args, pen=pg.mkPen(QColor(0, 255, 255)), **kwargs):
        super(LineROI, self).__init__(*args, pen=pen, **kwargs)
        self._update_state()

    def _update_state(self):
        self.width = self.size().y()
        self.length = self.size().x()
        self.rotation = self.angle()
        self.center_x = self.pos().x()
        self.center_y = self.pos().y()

    def movePoint(self, handle, pos, modifiers=Qt.KeyboardModifier(), finish=True, coords="parent"):
        super(LineROI, self).movePoint(handle, pos, modifiers, finish, coords)

        self._update_state()
        self.handleChanged()

    def mouseDragEvent(self, ev):
        super(LineROI, self).mouseDragEvent(ev)
        self._update_state()

    def paint(self, p, opt, widget):
        self.setSize(QPointF(self.length, self.width))
        self.setAngle(self.rotation)
        self.setPos(QPointF(self.center_x, self.center_y))
        super(LineROI, self).paint(p, opt, widget)

    def parameter(self) -> Parameter:
        if not self._param:
            self._param = parameterTypes.GroupParameter(
                name="Line ROI",
                children=[
                    parameterTypes.SimpleParameter(
                        title="Center X", name="center_x", value=self.center_x, type="float", units="px"
                    ),
                    parameterTypes.SimpleParameter(
                        title="Center Y", name="center_y", value=self.center_y, type="float", units="px"
                    ),
                    parameterTypes.SimpleParameter(
                        title="Rotation Angle", name="rotation", value=self.rotation, type="float", units="px"
                    ),
                    parameterTypes.SimpleParameter(
                        title="Length", name="length", value=self.length, type="float", units="px"
                    ),
                    parameterTypes.SimpleParameter(title="Width", name="width", value=self.width, type="float", units="px"),
                ],
            )

            self._param.sigTreeStateChanged.connect(self.valueChanged)

        return self._param

    def handleChanged(self):
        self.parameter().child("center_x").setValue(self.center_x)
        self.parameter().child("center_y").setValue(self.center_y)
        self.parameter().child("rotation").setValue(self.rotation)
        self.parameter().child("width").setValue(self.width)
        self.parameter().child("length").setValue(self.length)


class SegmentedRectROI(RectROI):
    def __init__(self, *args, **kwargs):
        self.segments_h = 2
        self.segments_v = 2
        super(SegmentedRectROI, self).__init__(*args, **kwargs)

    def parameter(self) -> Parameter:
        if not self._param:
            self._param = parameterTypes.GroupParameter(
                name="Rectangular ROI",
                children=[
                    parameterTypes.SimpleParameter(title="Width", name="width", value=self.width, type="float", units="px"),
                    parameterTypes.SimpleParameter(
                        title="Height", name="height", value=self.height, type="float", units="px"
                    ),
                    parameterTypes.SimpleParameter(
                        title="Horizontal Segments", name="segments_h", value=self.segments_h, type="int"
                    ),
                    parameterTypes.SimpleParameter(
                        title="Vertical Segments", name="segments_v", value=self.segments_v, type="int"
                    ),
                ],
            )

            self._param.sigTreeStateChanged.connect(self.valueChanged)

        return self._param

    def getLabelArray(self, arr, img=None):
        """
        Return the result of ROI.getArrayRegion() masked by the arc shape
        of the ROI. Regions outside the arc are set to 0.
        """
        w, h = arr.shape

        min_x = self.pos().x()
        min_y = self.pos().y()
        max_x = self.size().x() + min_x
        max_y = self.size().y() + min_y
        segment_bin_x = (max_x - min_x) / self.segments_h
        segment_bin_y = (max_y - min_y) / self.segments_v

        mask = np.zeros_like(arr)

        for i in range(self.segments_h):
            for j in range(self.segments_v):
                # generate an square max
                label_mask = np.fromfunction(
                    lambda x, y: (x + 0.5 > min_x + i * segment_bin_x)
                    & (x + 0.5 < min_x + (i + 1) * segment_bin_x)
                    & (y + 0.5 > min_y + j * segment_bin_y)
                    & (y + 0.5 < min_y + (j + 1) * segment_bin_y),
                    (w, h),
                )
                mask[label_mask] = 1 + i + j * self.segments_h

        return mask

    def paint(self, p, opt, widget):
        super(SegmentedRectROI, self).paint(p, opt, widget)

        min_x = self.pos().x()
        min_y = self.pos().y()
        max_x = self.size().x() + min_x
        max_y = self.size().y() + min_y
        segment_bin_x = (max_x - min_x) / self.segments_h
        segment_bin_y = (max_y - min_y) / self.segments_v

        self.currentPen.setStyle(Qt.DashLine)
        p.setPen(self.currentPen)

        for i in range(1, self.segments_h):
            p.drawLine(QPointF(1.0 / self.segments_h * i, 0), QPointF(1 / self.segments_h * i, 1))

        for j in range(1, self.segments_v):
            p.drawLine(QPointF(0, 1 / self.segments_v * j), QPointF(1, 1 / self.segments_v * j))

        self.currentPen.setStyle(Qt.SolidLine)


if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication

    qapp = QApplication([])
    import pyqtgraph as pg

    imageview = pg.ImageView()
    imageview.view.invertY(False)
    data = np.random.random((10, 10))
    imageview.setImage(data)

    # roi = ArcROI(center=(5, 5), radius=5)
    roi = SegmentedRectROI(pos=(0, 0), size=(10, 10))
    imageview.view.addItem(roi)

    imageview2 = pg.ImageView()
    imageview2.view.invertY(False)

    def showroi(*_, **__):
        imageview2.setImage(roi.getLabelArray(data, imageview.imageItem))

    roi.sigRegionChanged.connect(showroi)

    imageview.show()
    imageview2.show()
    qapp.exec_()



from typing import List
from qtpy.QtCore import QObject, QSettings
from .plugin import PluginType
from xicam import plugins
from pyqtgraph.parametertree import ParameterTree
from pyqtgraph.parametertree.parameterTypes import GroupParameter
import cloudpickle as pickle
from xicam.core import msg


class SettingsPlugin(QObject, PluginType):
    is_singleton = True
    needs_qt = False

    def __init__(self, icon, name, widget):
        super(SettingsPlugin, self).__init__()
        self.icon = icon
        self._name = name
        self._widget = widget

    @property
    def widget(self):
        return self._widget

    @widget.setter
    def widget(self, widget):
        self._widget = widget

    def name(self):
        return self._name

    def apply(self):
        ...

    def toState(self):
        self.apply()
        ...

    def fromState(self, state):
        ...

    def save(self):
        QSettings().setValue(self.name(), pickle.dumps(self.toState()))

    def restore(self):
        try:
            state = QSettings().value(self.name())
            if state != pickle.dumps(self.toState()):
                self.fromState(pickle.loads(state))
            # else:
            #     msg.logMessage(f"skipped restoring {self.name()}")
        except (AttributeError, TypeError, SystemError, KeyError, ModuleNotFoundError) as ex:
            # No settings saved
            msg.logError(ex)
            msg.logMessage(
                f"Could not restore settings for {self.name} plugin; re-initializing settings...", level=msg.WARNING
            )


class ParameterSettingsPlugin(GroupParameter, SettingsPlugin):
    def __init__(self, icon, name: str, paramdicts: List[dict], **kwargs):
        SettingsPlugin.__init__(self, icon, name, None)
        GroupParameter.__init__(self, name=name, type="group", children=paramdicts, **kwargs)
        self.restore()

    @property
    def widget(self):
        widget = ParameterTree()
        widget.setParameters(self, showTop=False)
        return widget

    def apply(self):
        pass

    def toState(self):
        self.apply()
        return self.saveState(filter="user")

    def fromState(self, state):
        self.restoreState(state, addChildren=False, removeChildren=False)



# Adapted from http://doc.qt.io/qt-5/qtwidgets-dialogs-configdialog-configdialog-cpp.html under BSD


# TODO Add remotes config
# TODO Add usage statistics config

from qtpy.QtCore import Qt, QSize
from qtpy.QtGui import QStandardItemModel, QStandardItem, QKeyEvent
from qtpy.QtWidgets import (
    QDialog,
    QApplication,
    QListView,
    QStackedWidget,
    QHBoxLayout,
    QVBoxLayout,
    QAbstractItemView,
    QDialogButtonBox,
)

from xicam.plugins import manager as pluginmanager, Filters
from xicam.core import msg


class ConfigRestorer:
    def __init__(self):
        # Restore Settings
        pluginmanager.attach(self.request_reload, Filters.COMPLETE)

    @staticmethod
    def request_reload():
        for plugin in pluginmanager.get_plugins_of_type("SettingsPlugin"):
            plugin.restore()
            plugin.save()


class ConfigDialog(QDialog):
    def __init__(self):
        super(ConfigDialog, self).__init__()

        self.needs_reload = True

        # Set size and position
        self.setGeometry(0, 0, 900, 550)
        frameGm = self.frameGeometry()
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        centerPoint = QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())

        self.contentsWidget = QListView()
        self.contentsWidget.setViewMode(QListView.IconMode)
        # self.contentsWidget.setIconSize(QSize(96, 84))
        self.contentsWidget.setMovement(QListView.Static)
        self.contentsWidget.setMaximumWidth(174)
        self.contentsWidget.setSpacing(12)
        self.contentsWidget.setSelectionMode(QAbstractItemView.SingleSelection)

        self.contentsModel = QStandardItemModel()
        self.contentsWidget.setModel(self.contentsModel)
        self.contentsWidget.selectionModel().currentChanged.connect(self.changePage)

        self.buttonboxWidget = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel | QDialogButtonBox.Apply  # | QDialogButtonBox.Help
        )
        self.buttonboxWidget.button(QDialogButtonBox.Ok).clicked.connect(self.ok)
        self.buttonboxWidget.button(QDialogButtonBox.Apply).clicked.connect(self.apply)
        self.buttonboxWidget.button(QDialogButtonBox.Cancel).clicked.connect(self.close)

        self.pagesWidget = QStackedWidget()

        horizontalLayout = QHBoxLayout()
        horizontalLayout.addWidget(self.contentsWidget)
        horizontalLayout.addWidget(self.pagesWidget, 1)

        mainLayout = QVBoxLayout()
        mainLayout.addLayout(horizontalLayout)
        # mainLayout.addStretch(1)
        mainLayout.addSpacing(12)
        mainLayout.addWidget(self.buttonboxWidget)

        self.setLayout(mainLayout)
        self.setWindowTitle("Config Dialog")

        # Set modality
        self.setModal(True)

        self.lastwidget = None

        # Restore Settings
        pluginmanager.attach(self.request_reload, Filters.COMPLETE)

    def request_reload(self):
        if self.isVisible():
            self.reload()
        else:
            self.needs_reload = True  # The restore will happen when shown

    def createIcons(self):
        self.contentsModel.clear()
        for plugin in pluginmanager.get_plugins_of_type("SettingsPlugin"):
            item = QStandardItem(plugin.icon, plugin.name())
            item.widget = plugin.widget
            item.setTextAlignment(Qt.AlignHCenter)
            item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            item.setSizeHint(QSize(136, 80))
            self.contentsModel.appendRow(item)

    def show(self):
        if self.lastwidget:
            self.pagesWidget.addWidget(self.lastwidget)
            self.pagesWidget.setCurrentWidget(self.lastwidget)
        if self.needs_reload:
            self.reload()
        super(ConfigDialog, self).show()

    def changePage(self, current, previous):
        if not current:
            current = previous
        current = self.contentsModel.itemFromIndex(current)
        self.pagesWidget.addWidget(current.widget)
        self.pagesWidget.setCurrentWidget(current.widget)
        self.lastwidget = current.widget

    def reload(self):
        self.createIcons()

    def ok(self):
        self._empty()
        self.apply()
        self.accept()

    def apply(self):
        for plugin in pluginmanager.get_plugins_of_type("SettingsPlugin"):
            plugin.save()

    def close(self):
        self._empty()
        self.reject()

    def _empty(self):
        """
        Disown all widget children (otherwise their c++ objects are force deleted when the dialog closes).
        Must be run in reverse to avoid index update errors
        """
        for i in reversed(range(self.pagesWidget.count())):
            self.pagesWidget.widget(i).setParent(None)

    def closeEvent(self, event):
        self.close()
        event.accept()

    def keyPressEvent(self, e: QKeyEvent):
        if e.key() != Qt.Key_Escape:
            super(ConfigDialog, self).keyPressEvent(e)
        else:
            self.close()



def test_lazyfield():
    import fabio
    from xicam.core.data import lazyfield

    class Handler(object):
        def __init__(self, path):
            self.path = path

        def __call__(self, *args, **kwargs):
            return fabio.open(self.path).data

    # l = lazyfield(Handler, '/home/rp/data/YL1031/YL1031__2m_00000.edf')
    # assert l.asarray() is not None

    # def test_data():
    #     from .. import data
    #     data.load_header(filenames='/home/rp/data/YL1031/YL1031__2m_00000.edf')


import os
import entrypoints
import numpy as np
from databroker.in_memory import BlueskyInMemoryCatalog


def test_ingestor(tmp_path):
    # test data
    data = np.random.random((1000, 1000))
    # write data to test edf
    edf_path = os.path.join(tmp_path, "test.edf")
    print("edf_path:", edf_path)
    EdfImage(data).write(edf_path)

    # get edf ingestor
    edf_ingestor = entrypoints.get_single("databroker.ingestors", "application/edf").load()

    # load data into catalog
    document = list(edf_ingestor([edf_path]))
    uid = document[0][1]["uid"]
    catalog = BlueskyInMemoryCatalog()
    # TODO -- change upsert signature to put start and stop as kwargs
    # TODO -- ask about more convenient way to get a BlueskyRun from a document generator
    catalog.upsert(document[0][1], document[-1][1], edf_ingestor, ([edf_path],), {})
    return catalog[uid]


if __name__ == "__main__":
    run = test_ingestor("/tmp")
    print(run)
    print(list(run.canonical(fill="yes")))
    print(run.primary.to_dask().values)



import pytest

from xicam.core import execution
from xicam.core.execution import localexecutor
from xicam.core.execution.workflow import Graph, Workflow
from xicam.plugins import OperationPlugin
from xicam.plugins.operationplugin import output_names, operation

from xicam.core.tests.workflow_fixtures import graph, sum_op, square_op, negative_op


# Note that this test relies on the xicam.plugins module
# TODO prevent adding circular links
execution.executor = localexecutor.LocalExecutor()





class TestGraph:
    def test_add_operation(self, graph, sum_op):
        graph.add_operation(sum_op)
        assert graph.operations == [sum_op]

    def test_add_operations(self, graph, sum_op, square_op):
        operations = [sum_op, square_op]
        graph.add_operations(*operations)
        assert graph.operations == operations

    def test_insert_operation(self, graph, sum_op, square_op):
        graph.insert_operation(0, sum_op)
        graph.insert_operation(0, square_op)
        assert graph.operations == [square_op, sum_op]

    def test_add_link(self, graph, sum_op, square_op):
        graph.add_operation(sum_op)
        graph.add_operation(square_op)
        graph.add_link(source=sum_op, dest=square_op, source_param="sum", dest_param="n")
        sum_inbound_links = graph.get_inbound_links(sum_op)
        sum_outbound_links = graph.get_outbound_links(sum_op)
        square_inbound_links = graph.get_inbound_links(square_op)
        square_outbound_links = graph.get_outbound_links(square_op)
        links = [("sum", "n")]
        assert sum_inbound_links[square_op] == []
        assert sum_outbound_links[square_op] == links
        assert square_inbound_links[sum_op] == links
        assert square_outbound_links[sum_op] == []

    def test_add_multiple_links(self, graph, sum_op, square_op):
        graph.add_operation(sum_op)
        graph.add_operation(square_op)
        import math

        def my_sqrt(num):
            return math.sqrt(num)

        sqrt_op = operation(my_sqrt, output_names=("sqrt"))()
        graph.add_operation(sqrt_op)
        # sum -> square -> sqrt
        #   \               |
        #    -------->------
        graph.add_link(source=sum_op, dest=square_op, source_param="sum", dest_param="n")
        graph.add_link(source=square_op, dest=sqrt_op, source_param="square", dest_param="num")
        graph.add_link(source=sum_op, dest=sqrt_op, source_param="sum", dest_param="num")
        # sum -> square
        assert graph.get_inbound_links(sum_op)[square_op] == []
        assert graph.get_outbound_links(sum_op)[square_op] == [("sum", "n")]
        assert graph.get_inbound_links(square_op)[sum_op] == [("sum", "n")]
        assert graph.get_outbound_links(square_op)[sum_op] == []
        # square -> sqrt
        assert graph.get_inbound_links(square_op)[sqrt_op] == []
        assert graph.get_outbound_links(square_op)[sqrt_op] == [("square", "num")]
        assert graph.get_inbound_links(sqrt_op)[square_op] == [("square", "num")]
        assert graph.get_outbound_links(sqrt_op)[square_op] == []
        # sum -> sqrt
        assert graph.get_inbound_links(sum_op)[sqrt_op] == []
        assert graph.get_outbound_links(sum_op)[sqrt_op] == [("sum", "num")]
        assert graph.get_inbound_links(sqrt_op)[sum_op] == [("sum", "num")]
        assert graph.get_outbound_links(sqrt_op)[sum_op] == []

    def test_add_link_bad_source(self, graph, sum_op):
        with pytest.raises(TypeError):
            graph.add_link("bad", sum_op, "", "")

    def test_add_link_bad_dest(self, graph, sum_op):
        with pytest.raises(TypeError):
            graph.add_link(sum_op, "bad", "", "")

    def test_add_link_missing_source(self, graph, sum_op, square_op):
        graph.add_operation(square_op)
        with pytest.raises(ValueError):
            graph.add_link(sum_op, square_op, "sum", "n")

    def test_add_link_messing_dest(self, graph, sum_op, square_op):
        graph.add_operation(sum_op)
        with pytest.raises(ValueError):
            graph.add_link(sum_op, square_op, "sum", "n")

    def test_add_link_missing_source_param(self, graph, sum_op, square_op):
        with pytest.raises(ValueError):
            graph.add_link(source=sum_op, dest=square_op, source_param="sum", dest_param="dne")

    def test_add_link_missing_dest_param(self, graph, sum_op, square_op):
        with pytest.raises(ValueError):
            graph.add_link(source=sum_op, dest=square_op, source_param="dne", dest_param="n")

    def test_remove_link(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        graph.add_link(sum_op, square_op, source_param="sum", dest_param="n")
        graph.remove_link(sum_op, square_op, "sum", "n")
        assert len(graph.get_inbound_links(sum_op)) == 0 and len(graph.get_inbound_links(square_op)) == 0
        assert len(graph.get_outbound_links(sum_op)) == 0 and len(graph.get_outbound_links(square_op)) == 0

    def test_remove_link_no_links(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        with pytest.raises(ValueError):
            graph.remove_link(sum_op, square_op, "sum", "n")

    def test_get_inbound_links(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        assert graph.get_inbound_links(sum_op) == {}
        assert graph.get_inbound_links(square_op) == {sum_op: [("sum", "n")]}

    def test_get_inbound_links_empty(self, graph, sum_op):
        assert graph.get_inbound_links(sum_op) == {}

    def test_get_outbound_links(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        assert graph.get_outbound_links(square_op) == {}
        assert graph.get_outbound_links(sum_op) == {square_op: [("sum", "n")]}

    def test_get_outbound_links_empty(self, graph, sum_op):
        assert graph.get_outbound_links(sum_op) == {}

    def test_clear_operation_links_first(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        graph.add_link(square_op, negative_op, "square", "num")
        graph.clear_operation_links(sum_op)
        assert graph.links() == [(square_op, negative_op, "square", "num")]

    def test_clear_operation_links_middle(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        graph.add_link(square_op, negative_op, "square", "num")
        graph.clear_operation_links(square_op)
        assert graph.links() == []

    def test_clear_operation_links_end(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        graph.add_link(square_op, negative_op, "square", "num")
        graph.clear_operation_links(negative_op)
        assert graph.links() == [(sum_op, square_op, "sum", "n")]

    def test_clear_operation_links_empty(self, graph, sum_op):
        graph.clear_operation_links(sum_op)
        assert graph.links() == []
        assert graph.get_outbound_links(sum_op) == {}
        assert graph.get_inbound_links(sum_op) == {}

    def test_clear_operation_links_bad_arg(self, graph):
        with pytest.raises(TypeError):
            graph.clear_operation_links("this should be an operation")

    def test_clear_operation_links_unlinked_operation(self, graph, sum_op, square_op):
        # TODO should this raise an exception?
        graph.add_operations(sum_op, square_op)
        graph.clear_operation_links(sum_op)
        assert graph.links() == []

    def test_clear_links_empty(self, graph):
        # TODO should this raise an exception
        graph.clear_links()
        assert graph.links() == []

    def test_clear_links(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        graph.add_link(square_op, negative_op, "square", "num")
        graph.clear_links()
        assert graph.links() == []

    def test_clear_operations_empty(self, graph):
        graph.clear_operations()
        assert graph.operations == []

    def test_clear_operations(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        graph.clear_operations()
        assert graph.operations == []

    def test_remove_operation_empty(self, graph, sum_op):
        with pytest.raises(ValueError):
            graph.remove_operation(sum_op)

    def test_remove_operation(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        return_value = graph.remove_operation(sum_op)
        assert graph.operations == [square_op]
        assert return_value is None  # no return by default

    def test_remove_operation_linked(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        graph.add_link(square_op, negative_op, "square", "num")
        graph.remove_operation(sum_op)
        assert graph.operations == [square_op, negative_op]
        assert graph.links() == [(square_op, negative_op, "square", "num")]

    def test_remove_operation_not_remove_orphan_links(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        to_remove = graph.remove_operation(square_op, remove_orphan_links=False)
        assert graph.operations == [sum_op]
        assert to_remove == (graph.get_inbound_links(square_op), graph.get_outbound_links(square_op))

    def test_remove_operation_bad_arg(self, graph):
        with pytest.raises(TypeError):
            graph.remove_operation("this should be an operation")

    def test_remove_operation_not_in_graph(self, graph, sum_op):
        with pytest.raises(ValueError):
            graph.remove_operation(sum_op)

    def test_as_dask_graph(self, graph, sum_op, square_op, negative_op):
        # Connect sum_op to square_op; don't connect negative_op
        graph.add_operations(sum_op, square_op, negative_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        dask_graph, end_ids = graph.as_dask_graph()

        # Should look like:
        # { "0": (<sum_op>,), "1": (<square_op>, "0"), "2": (<negative_op>,) }
        sum_wrapper = dask_graph["0"]
        square_wrapper = dask_graph["1"]
        negative_wrapper = dask_graph["2"]
        assert len(sum_wrapper) == 1
        assert sum_wrapper[0].node is sum_op
        assert len(square_wrapper) == 2
        assert square_wrapper[0].node is square_op
        assert square_wrapper[1] == "0"
        assert len(negative_wrapper) == 1
        assert negative_wrapper[0].node is negative_op

        # Both square_op and negative_op should be end nodes
        assert sorted(end_ids) == sorted(["1", "2"])

    def test_as_dask_graph_multiple_links(self, graph, sum_op, square_op, negative_op):
        def my_func(x: int, y: int) -> (int, int):
            return y, x

        # Connect sum_op to square_op.
        # Connect sum_op to my_op's x, square_op to my_op's y.
        # Leave negative_op unconnected
        my_op = operation(my_func, output_names=("y", "x"))()
        graph.add_operations(sum_op, square_op, negative_op, my_op)
        graph.add_link(sum_op, square_op, "sum", "n")
        graph.add_link(sum_op, my_op, "sum", "x")
        graph.add_link(square_op, my_op, "square", "y")
        dask_graph, end_ids = graph.as_dask_graph()

        # Should look like:
        sum_wrapper = dask_graph["0"]
        square_wrapper = dask_graph["1"]
        negative_wrapper = dask_graph["2"]
        my_wrapper = dask_graph["3"]

        # sum_op has no dependent nodes (no ops connect into it)
        assert len(sum_wrapper) == 1
        assert sum_wrapper[0].node is sum_op

        # square_op has 1 dependent node, takes sum_op's output
        assert len(square_wrapper) == 2
        assert square_wrapper[0].node is square_op

        # negative_op has no dependent nodes; is unconnected
        assert len(negative_wrapper) == 1
        assert negative_wrapper[0].node is negative_op

        # my_op has two dependent nodes; sum_op and square_op connect to its inputs
        assert len(my_wrapper) == 3
        assert my_wrapper[0].node is my_op
        assert my_wrapper[1] == "0"  # sum_op
        assert my_wrapper[2] == "1"  # square_op

        # negative_op, and my_op should be end nodes
        assert sorted(end_ids) == sorted(["2", "3"])

    def test_as_dask_graph_empty(self, graph):
        # Empty graph, no end nodes
        assert graph.as_dask_graph() == ({}, [])

    def test_as_dask_graph_no_links(self, graph, sum_op):
        graph.add_operation(sum_op)
        dask_graph, end_ids = graph.as_dask_graph()
        assert len(dask_graph["0"]) == 1
        assert dask_graph["0"][0].node is sum_op
        assert end_ids == ["0"]

    def test_operations(self, graph, sum_op):
        graph.add_operation(sum_op)
        assert graph.operations == [sum_op]

    def test_operations_empty(self, graph):
        assert graph.operations == []

    def test_links(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        link = (sum_op, square_op, "sum", "n")
        graph.add_link(*link)
        assert graph.links() == [link]

    def test_links_multiple(self, graph, sum_op):
        def my_func(x: int, y: int) -> (int, int):
            return y, x

        my_op = operation(my_func, output_names=("y", "x"))()
        graph.add_operations(sum_op, my_op)
        link1 = (my_op, sum_op, "y", "n1")
        link2 = (my_op, sum_op, "x", "n2")
        graph.add_link(*link1)
        graph.add_link(*link2)
        assert graph.links() == [link1, link2]

    def test_links_empty(self, graph):
        assert graph.links() == []

    def test_operation_links_bad_arg(self, graph):
        with pytest.raises(TypeError):
            graph.operation_links("this is not an operation")

    def test_operation_links_not_in_graph(self, graph, sum_op):
        with pytest.raises(ValueError):
            graph.operation_links(sum_op)

    def test_operation_links(self, graph, sum_op, square_op):
        graph.add_operations(sum_op, square_op)
        link = (sum_op, square_op, "sum", "n")
        graph.add_link(*link)
        assert graph.operation_links(sum_op) == [link]
        assert graph.operation_links(square_op) == []

    def test_operation_links_multiple(self, graph, sum_op, square_op, negative_op):
        def my_func(x: int, y: int) -> (int, int):
            return y, x

        my_op = operation(my_func, output_names=("y", "x"))()
        graph.add_operations(sum_op, square_op, negative_op, my_op)
        link1 = (my_op, sum_op, "y", "n1")
        link2 = (my_op, sum_op, "x", "n2")
        link3 = (sum_op, square_op, "sum", "n")
        link4 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        graph.add_link(*link3)
        graph.add_link(*link4)
        assert graph.operation_links(my_op) == [link1, link2]
        assert graph.operation_links(sum_op) == [link3]
        assert graph.operation_links(negative_op) == []

    def test_set_disabled_default_no_links(self, graph, sum_op):
        graph.add_operation(sum_op)
        return_value = graph.set_disabled(sum_op)
        assert graph.disabled(sum_op) is True
        assert return_value == []

    # TODO parameterize these tests
    def test_set_disabled_default_with_links(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        link1 = (sum_op, square_op, "sum", "n")
        link2 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        return_value = graph.set_disabled(sum_op)
        assert graph.disabled(sum_op) is True
        assert graph.links() == [link2]
        assert return_value == []

    def test_set_disabled_remove_false(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        link1 = (sum_op, square_op, "sum", "n")
        link2 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        return_value = graph.set_disabled(sum_op, remove_orphan_links=False)
        assert graph.disabled(sum_op) is True
        assert graph.links() == [link1, link2]
        assert return_value == graph.operation_links(sum_op)

    def test_set_disabled_value_false(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        link1 = (sum_op, square_op, "sum", "n")
        link2 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        return_value = graph.set_disabled(sum_op, value=False)
        assert graph.disabled(sum_op) is False
        assert graph.links() == [link1, link2]
        assert return_value == []

    def test_set_disabled_value_and_remove_false(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        link1 = (sum_op, square_op, "sum", "n")
        link2 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        return_value = graph.set_disabled(sum_op, value=False, remove_orphan_links=False)
        assert graph.disabled(sum_op) is False
        assert graph.links() == [link1, link2]
        assert return_value == []

    def test_toggle_disabled_default(self, graph, sum_op):
        graph.add_operation(sum_op)
        return_value = graph.toggle_disabled(sum_op)
        assert graph.disabled(sum_op) is True
        assert return_value == []
        return_value = graph.toggle_disabled(sum_op)
        assert graph.disabled(sum_op) is False
        assert return_value == []

    def test_toggle_disabled_with_links(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        link1 = (sum_op, square_op, "sum", "n")
        link2 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        return_value = graph.toggle_disabled(sum_op)
        assert graph.disabled(sum_op) is True
        assert return_value == []
        assert graph.links() == [link2]
        return_value = graph.toggle_disabled(sum_op)
        assert graph.disabled(sum_op) is False
        assert return_value == []
        assert graph.links() == [link2]

    def test_toggle_disabled_remove_false(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        link1 = (sum_op, square_op, "sum", "n")
        link2 = (square_op, negative_op, "square", "num")
        graph.add_link(*link1)
        graph.add_link(*link2)
        return_value = graph.toggle_disabled(sum_op, remove_orphan_links=False)
        assert graph.disabled(sum_op) is True
        assert return_value == graph.operation_links(sum_op)
        assert graph.links() == [link1, link2]
        return_value = graph.toggle_disabled(sum_op, remove_orphan_links=False)
        assert graph.disabled(sum_op) is False
        assert return_value == []
        assert graph.links() == [link1, link2]

    def test_auto_connect_all_only_matching_names(self, graph):
        def my_increment(n: int) -> int:
            return n + 1

        def my_decrement(increment: float) -> float:
            return increment - 1

        increment_op = operation(my_increment, output_names=("increment",))()
        decrement_op = operation(my_decrement, output_names=("end_result",))()
        graph.add_operations(increment_op, decrement_op)
        graph.auto_connect_all()
        assert graph.links() == [(increment_op, decrement_op, "increment", "increment")]

    def test_auto_connect_all_only_matching_types(self, graph):
        def my_increment(n: int) -> int:
            return n + 1

        def my_decrement(m: int) -> int:
            return m - 1

        increment_op = operation(my_increment, output_names=("increment",))()
        decrement_op = operation(my_decrement, output_names=("decrement",))()
        graph.add_operations(increment_op, decrement_op)
        graph.auto_connect_all()
        assert graph.links() == []

    # TODO better test with multiple viable auto links, then prefer the one with matching type as well
    def test_auto_connect_all_matching_names_and_types(self, graph):
        def my_increment(n: int) -> int:
            return n + 1

        def my_decrement(increment: int) -> int:
            return increment - 1

        increment_op = operation(my_increment, output_names=("increment",))()
        decrement_op = operation(my_decrement, output_names=("end_result",))()
        graph.add_operations(increment_op, decrement_op)
        graph.auto_connect_all()
        assert graph.links() == [(increment_op, decrement_op, "increment", "increment")]

    def test_auto_connect_all_none_matching(self, graph, sum_op, square_op, negative_op):
        graph.add_operations(sum_op, square_op, negative_op)
        graph.auto_connect_all()
        assert graph.links() == []


class TestWorkflow:
    # TODO: test callback_slot
    # TODO: test finished_slot
    # TODO: test except_slot
    # TODO: test execute_all

    def test_execute_no_operations(self):
        workflow = Workflow("Test Workflow")
        results = workflow.execute().result()
        assert results == ({},)

    def test_execute_operation_no_default_no_value(self, sum_op):
        # not sure how to test this....
        def handle_exception(exception):
            with pytest.raises(TypeError):
                raise exception

        workflow = Workflow()
        workflow.add_operation(sum_op)
        results = workflow.execute(except_slot=handle_exception).result()
        print(results)

    def test_execute_operation_no_default(self, sum_op):
        workflow = Workflow()
        workflow.add_operation(sum_op)
        results = workflow.execute(n1=10, n2=5).result()
        assert results == ({"sum": 15},)

    def test_execute_operation_default_input_value(self):
        @operation
        @output_names("doubled")
        def double_op(x=10):
            return x * 2

        workflow = Workflow()
        workflow.add_operation(double_op)
        results = workflow.execute().result()
        assert results == ({"doubled": 20},)

    def test_execute_no_links_TODO(self, sum_op, square_op, negative_op):
        # do the input names have to match in this case (more than one entry op)
        operations = [sum_op, square_op, negative_op]
        workflow = Workflow(name="test", operations=operations)
        results = workflow.execute(n1=3, n2=-3).result()
        assert results == ({"sum": 0})

    def test_execute_no_links(self):
        @operation
        @output_names("doubled")
        def double_op(n):
            return n * 2

        @operation
        @output_names("tripled")
        def triple_op(n):
            return n * 3

        workflow = Workflow()
        workflow.add_operations(double_op, triple_op)
        results = workflow.execute(n=10).result()
        assert len(results) == 2
        assert {"doubled": 20} in results
        assert {"tripled": 30} in results

    def test_execute(self, sum_op, square_op, negative_op):
        workflow = Workflow()
        workflow.add_operations(sum_op, square_op, negative_op)
        workflow.add_link(sum_op, square_op, "sum", "n")
        workflow.add_link(square_op, negative_op, "square", "num")
        results = workflow.execute(n1=2, n2=5).result()
        assert results == ({"negative": -49},)

    def test_execute_synchronous(self, sum_op, square_op, negative_op):
        workflow = Workflow()
        workflow.add_operations(sum_op, square_op, negative_op)
        workflow.add_link(sum_op, square_op, "sum", "n")
        workflow.add_link(square_op, negative_op, "square", "num")
        results = workflow.execute_synchronous(n1=2, n2=5)
        assert results == ({"negative": -49},)

    def test_execute_synchronous_no_links_not_enough_kwargs(self, sum_op, square_op, negative_op):
        # TODO -- expect exception
        workflow = Workflow()
        workflow.add_operations(sum_op, square_op, negative_op)
        results = workflow.execute_synchronous(n1=2, n2=5)
        assert results == ({"negative": -49},)

    def test_execute_synchronous_no_links(self, sum_op, square_op, negative_op):
        workflow = Workflow()
        workflow.add_operations(sum_op, square_op, negative_op)
        # n1, n2 -- inputs to sum_op;
        # n -- input to square_op
        # num -- input to negative_op
        results = workflow.execute_synchronous(n1=2, n2=5, n=10, num=50)
        assert len(results) == 3
        assert {"sum": 7} in results
        assert {"square": 100} in results
        assert {"negative": -50} in results

    def test_execute_all(self, sum_op, square_op, negative_op):
        workflow = Workflow()
        workflow.add_operations(sum_op, square_op, negative_op)
        workflow.add_link(sum_op, square_op, "sum", "n")
        workflow.add_link(square_op, negative_op, "square", "num")
        n1_values = [1, 3, 5]
        n2_values = [2, 4, 6]
        # TODO -- we are only getting one result, should get three (3 pairs of n1/n2).
        results = list(workflow.execute_all(n1=n1_values, n2=n2_values).result())
        assert results == [{"negative": (1 + 2) ** 2 * -1}, {"negative": (3 + 4) ** 2 * -1}, {"negative": (5 + 6) ** 2 * -1}]

    def test_fill_kwargs(self):
        assert False

    def test_validate(self):
        assert False

    def test_notify_no_observers(self):
        workflow = Workflow()
        workflow.notify()

    def test_notify(self):
        self.flag = False

        def observer():
            self.flag = not self.flag

        workflow = Workflow()
        workflow.attach(observer)
        workflow.notify()
        assert self.flag is True
